"""
FortiClient EMS monitor module
"""

import logging
from datetime import datetime
from typing import Any, Dict

from fotoobo.fortinet.forticlientems import FortiClientEMS
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def connections(host: str) -> Result[Dict[str, Any]]:
    """
    Get connections information from FortiClient EMS.

    The data returned is the raw data part from the EMS response. Additionally, the data is enriched
    with calculated and interesting values. All the enriched values are in the "fotoobo" in the
    returned dict.

    Args:
        host: FortiClient EMS host defined in the inventory

    Returns:
        Result
    """
    result = Result[Dict[str, Any]]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    ems.login()
    response = ems.api("get", "/endpoints/connection/donut")
    data: Dict[str, Any] = {"data": list(response.json()["data"]), "fotoobo": {}}

    for item in data["data"]:
        data["fotoobo"][item["token"]] = item["value"]

    result.push_result(host, data)
    return result


def endpoint_management_status(host: str) -> Result[Dict[str, Any]]:
    """
    Get management information about endpoints registered in FortiClient EMS.

    The data returned is the raw data part from the EMS response. Additionally, the data is enriched
    with calculated and interesting values. All the enriched values are in the key "fotoobo" in the
    returned dict:

        - {{ fotoobo.managed }}     Amount of managed endpoints
        - {{ fotoobo.unmanaged }}   Amount of unmanaged endpoints

    Args:
        host: FortiClient EMS host defined in the inventory

    Returns:
        Result
    """
    result = Result[Dict[str, Any]]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    ems.login()
    response = ems.api("get", "/endpoints/management/donut")
    data = {"data": dict(response.json())["data"]}
    managed = unmanaged = 0

    for item in data["data"]:
        if item["token"] == "managed":
            managed = item["value"]
            log.debug("Management: managed: '%s'", managed)

        if item["token"] == "unmanaged":
            unmanaged = item["value"]
            log.debug("Management: unmanaged: '%s'", unmanaged)

    data["fotoobo"] = {"managed": managed, "unmanaged": unmanaged}
    result.push_result(host, data)
    return result


def endpoint_online_outofsync(host: str) -> Result[Dict[str, Any]]:
    """
    Get amount of FortiClient EMS devices which are online but policy not in sync.

    The data returned is a dict with a key "fotoobo" with the relevant data. In this utility only
    one key value pair is present.

        - {{ fotoobo.outofsync }}   Amount of managed endpoints which are online but their policy
            is not in synch with FortiClient EMS

    Args:
        host: FortiClient EMS host defined in the inventory

    Returns:
        Result
    """
    result = Result[Dict[str, Any]]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    ems.login()
    response = ems.api(
        "get", "/endpoints/index?offset=0&count=1&connection=online&status=outofsync"
    )
    data = {"fotoobo": {"outofsync": response.json()["data"]["total"]}}
    log.debug("Endpoints outofsync: '%s'", data["fotoobo"]["outofsync"])
    result.push_result(host, data)
    return result


def endpoint_os_versions(host: str) -> Result[Dict[str, Dict[str, Any]]]:
    """
    Get management information about endpoints registered in FortiClient EMS.

    The data returned is the raw data part from the EMS response. Additionally, the data is enriched
    with calculated and interesting values. All the enriched values are in the key "fotoobo" in the
    returned dict:

        - {{ fotoobo.fctversionlinux }}     Amount of managed linux endpoints
        - {{ fotoobo.fctversionmac }}       Amount of managed mac endpoints
        - {{ fotoobo.fctversionwindows }}   Amount of managed windows endpoints

    Args:
        host: FortiClient EMS host defined in the inventory

    Returns:
        Result
    """
    result = Result[Dict[str, Dict[str, Any]]]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    ems.login()
    data: Dict[str, Any] = {"data": {}, "fotoobo": {}}

    for fctversion_os in ["fctversionwindows", "fctversionmac", "fctversionlinux"]:
        response = ems.api("get", f"/endpoints/{fctversion_os}/donut")
        data["data"][fctversion_os] = dict(response.json())["data"]
        count = sum(item["value"] for item in data["data"][fctversion_os])
        data["fotoobo"][fctversion_os] = count

    result.push_result(host, data)
    return result


def system(host: str) -> Result[Dict[str, Any]]:
    """
    Get system information from FortiClient EMS.

    Args:
        host: FortiClient EMS host defined in inventory

    Returns:
        Result
    """
    result = Result[Dict[str, Any]]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    ems.login()

    # get EMS serial number (just for debug logging and because it's possible)
    response = ems.api("get", "/system/serial_number")
    log.debug("Serial number: '%s' (from /system/serial_number)", response.json()["data"])

    # get EMS system info
    response = ems.api("get", "/system/info/")
    log.debug("Serial number: '%s' (from /system/info/)", response.json()["data"]["license"]["sn"])
    result.push_result(host, dict(response.json()["data"]))
    return result


def license(host: str) -> Result[Dict[str, Any]]:  # pylint: disable=redefined-builtin
    """
    Get license information from FortiClient EMS.

    The data returned is the raw data part from the EMS response. Additionally, the data is enriched
    with calculated and interesting values. All the enriched values are in the key "fotoobo" in the
    returned dict:

        - {{ fotoobo.[MODULE]_usage }}
            License usage in percent by MODULE where MODULE is one of:
                - fabric_agent
                - sandbox_cloud
                - chromebook
                - ztna
                - epp
                - sase
                - ztna_user
                - epp_user
                - sase_user
                - vpn_only
        - {{ fotoobo.license_expiry_days }}     Days until your FortiClient EMS license expires.
            Be aware that it only shows the expiry of the "fabric_agent" fabric_agent as for now.
            If you wish more granular view this utility has to be improved accordingly.

    Args:
        host: FortiClient EMS host defined in the inventory

    Returns:
        Result
    """
    result = Result[Dict[str, Any]]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    ems.login()
    response = ems.api("get", "/license/get")
    data = {}
    data["data"] = dict(response.json()["data"])
    license_expiry_days = 0

    for lic in data["data"]["licenses"]:
        log.debug("License expiry for '%s': '%s'", lic["type"], lic["expiry_date"])
        license_expiry = datetime.strptime(lic["expiry_date"], "%Y-%m-%dT%H:%M:%S")
        expiry_days = int((license_expiry - datetime.now()).days)

        if expiry_days < license_expiry_days or license_expiry_days == 0:
            license_expiry_days = expiry_days

    log.debug("Days to license expiry: '%s'", license_expiry_days)
    data["fotoobo"] = {}
    data["fotoobo"]["license_expiry_days"] = license_expiry_days

    for key in data["data"]["seats"]:
        if data["data"]["seats"][key] > 0:
            log.debug(f"{key} license count: '%s'", data["data"]["seats"][key])
            log.debug(f"{key} license used: '%s'", data["data"]["used"][key])
            license_usage = int(100 / data["data"]["seats"][key] * data["data"]["used"][key])
            log.debug(f"{key} license usage : '%s%%'", license_usage)
            data["fotoobo"][key + "_usage"] = license_usage

    result.push_result(host, data)
    return result



"""
The fotoobo get commands
"""

import logging

import typer
from rich import print as rich_print
from rich.panel import Panel

from fotoobo.helpers import cli_path
from fotoobo.tools import get

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fotoobo get command callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def inventory() -> None:
    """
    Print a summary over your fotoobo inventory.
    """
    result = get.inventory()
    result.print_result_as_table(title="fotoobo inventory", headers=["Device", "Hostname", "Type"])


@app.command()
def version(
    verbose: bool = typer.Option(
        False, "-v", help="Verbose output (also show the most important modules)."
    ),
) -> None:
    """
    Print the fotoobo version.
    """
    result = get.version(verbose)

    if verbose:
        out_list = [
            {
                "module": "[bold]fotoobo[/]",
                "version": f"[bold]{result.get_result('version')[0]['version']}[/]",
            }
        ]
        out_list += result.get_result("version")[1:]

    else:
        out_list = result.get_result("version")

    result.print_table_raw(out_list, headers=[], auto_header=False, title="fotoobo version")


@app.command()
def commands() -> None:
    """
    Print the fotoobo commands structure.
    """
    result = get.commands()
    tree = result.get_result("commands")
    rich_print(
        Panel(tree, border_style="#FF33BB", title="cli commands structure", title_align="right")
    )



"""
The fotoobo get utility
"""

import importlib.metadata
import logging
from typing import Dict, List

from rich.text import Text
from rich.tree import Tree

from fotoobo import __version__
from fotoobo.helpers.cli import walk_cli_info
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def inventory() -> Result[Dict[str, str]]:
    """
    Get the fotoobo inventory

    Returns:
        Result
    """
    log.debug("Print fotoobo inventory information")
    result = Result[Dict[str, str]]()
    _inventory = Inventory(config.inventory_file)

    for host, data in _inventory.assets.items():
        result.push_result(host, {"hostname": data.hostname, "type": data.type})

    return result


def version(verbose: bool = False) -> Result[List[Dict[str, str]]]:
    """
    Get the fotoobo version

    Args:
        verbose: Whether we want verbose output
    """
    log.debug("Print fotoobo version information: '%s'", __version__)
    result = Result[List[Dict[str, str]]]()
    versions = [{"module": "fotoobo", "version": __version__}]

    if verbose:
        modules = ["jinja2", "PyYAML", "requests", "rich", "typer"]

        for module in modules:
            versions.append({"module": module, "version": importlib.metadata.version(module)})

    result.push_result("version", versions)
    return result


def commands() -> Result[Tree]:
    """Get the fotoobo cli commands

    Walk through the typer cli app and return its commands as a beautiful rich tree. The commands
    are sorted in alphabetical order
    """
    result = Result[Tree]()
    result.push_result(
        "commands",
        walk_cli_info(
            config.cli_info["command"],
            Tree(Text().append(config.cli_info["info_name"], style="bold cyan")),
        ),
    )

    return result



"""
FortiGate get version utility
"""

import concurrent.futures
import logging
from typing import Optional, Tuple

from rich.progress import Progress

from fotoobo.exceptions import GeneralError, GeneralWarning
from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def version(host: Optional[str] = None) -> Result[str]:
    """FortiGate get version.

    Get the version(s) of one ore more FortiGates.

    Args:
        host: The host from the inventory to get the version. If you omit host, it will run over
              all FortiGates in the inventory.

    Returns:
        The Result object with all the results
    """

    def _get_single_version(name: str, fgt: FortiGate) -> Tuple[str, str]:
        """Get the version from a FortiGate.

        This private method is used for multithreading. It only queries one single FortiGate for its
        version number status and returns it.

        Args:
            name: The name of the FortiGate (as defined in the inventory)
            fgt:  The FortiGate object to query

        Returns:
            name:   The name of the FortiGate (as defined in the inventory)
            status: The HA status of the FortiGate (fgt)
        """
        log.debug("Getting FortiGate version for '%s'", name)
        try:
            fortigate_version = fgt.get_version()

        except (GeneralWarning, GeneralError) as exception:
            fortigate_version = f"unknown due to {exception.message}"

        return name, fortigate_version

    inventory = Inventory(config.inventory_file)
    fgts = inventory.get(host, "fortigate")
    result = Result[str]()

    with Progress() as progress:
        task = progress.add_task("getting FortiGate versions...", total=len(fgts))
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []
            for name, fgt in fgts.items():
                futures.append(executor.submit(_get_single_version, name, fgt))

            for future in concurrent.futures.as_completed(futures):
                name, fortigate_version = future.result()
                result.push_result(name, fortigate_version)
                progress.update(task, advance=1)

    return result



"""
This is the Fortinet abstract base class (ABC) which is used to define some global and generic
variables and methods.
"""

import logging
from abc import ABC, abstractmethod
from time import time
from typing import Any, Dict, Optional, Union

import requests
import urllib3

from fotoobo.exceptions import APIError, GeneralError

log = logging.getLogger("fotoobo")


class Fortinet(ABC):
    """
    This is the Fortinet abstract base class. All other Fortinet product classes should inherit
    from this class. If there are methods which have to be defined in every subclass it has to be
    defined here with the abstractmethod decorator.
    """

    # Use the ALLOWED_HTTP_METHODS class constant to define the supported HTTP methods. By default
    # we should support GET and POST but you may override this list of supported methods in every
    # subclass. Treat this setting as a constant which must not be redefined during runtime.
    ALLOWED_HTTP_METHODS = ["GET", "POST"]

    def __init__(self, hostname: str, **kwargs: Any) -> None:
        """
        Set some initial parameters for the Fortinet super class.

        It also initializes a requests session. If you're making several requests to the same host,
        the underlying TCP connection will be reused, which can result in a significant performance
        increase. (https://docs.python-requests.org/en/master/user/advanced/)

        Args:
            hostname: The hostname of the Fortinet device to connect to

        Keyword Args:
            https_port: The tcp port number to connect to the https api
                If you do not specify a port number it is set to 443 by default.
            proxy: The Proxy server to use to connect to the Fortinet device
                If you need to connect to your Fortinet device through a proxy server you can
                set it here as as string. If needed you may append the proxy server port with a
                column to the proxy server. e.g. "proxy.local:8000".
            ssl_verify: Enable or disable SSL certificate check
                When ssl_verify is enabled you have to install a trusted SSL certificate onto
                the device you wish to connect to. If you set ssl_verify to false it will also
                disable the warnings in urllib3. This prevents unwanted SSL warnings to be
                logged.
            timeout: Connection timeout in seconds
        """
        self.api_url: str = ""
        self.hostname: str = hostname
        self.https_port: int = kwargs.get("https_port", 443)
        self.session = requests.Session()
        self.session.trust_env = False
        self.session.proxies: Dict[str, Any] = {"http": None, "https": None}  # type: ignore

        if proxy := kwargs.get("proxy", ""):
            self.session.proxies = {"http": f"{proxy}", "https": f"{proxy}"}

        self.ssl_verify: Union[bool, str] = kwargs.get("ssl_verify", True)
        if not self.ssl_verify:
            urllib3.disable_warnings(category=urllib3.exceptions.InsecureRequestWarning)

        self.timeout = kwargs.get("timeout", 3)
        self.type: str = ""

    def api(  # pylint: disable=too-many-arguments
        self,
        method: str,
        url: str = "",
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        payload: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
    ) -> requests.models.Response:
        """
        API request to a Fortinet device.

        Args:
            method:     HTTP request method
            url:        Rest API URL to request data from
            headers:    Dictionary with headers (if needed)
            params:     Dictionary with parameters (if needed)
            payload:    JSON body for post requests (if needed)
            timeout:    The requests read timeout

        Returns:
            Response from the request
        """
        full_url = f"{self.api_url}/{url.strip('/')}".strip("/")
        timeout = timeout or self.timeout
        start = time()

        if method.upper() not in self.ALLOWED_HTTP_METHODS:
            error = f"HTTP method '{method.upper()}' is not implemented"
            log.error(error)
            raise NotImplementedError(error)

        try:
            response: requests.Response = getattr(self.session, method.lower())(
                full_url,
                headers=headers,
                json=payload,
                params=params,
                timeout=timeout,
                verify=self.ssl_verify,
            )

        except requests.exceptions.SSLError as err:
            log.debug(err)
            error = "Unknown SSL error"

            try:
                if (
                    err.args[0].reason.args[0].verify_message
                    == "unable to get local issuer certificate"
                ):
                    error = "Unable to get local issuer certificate"

            except (AttributeError, IndexError):
                pass

            raise GeneralError(f"{error} ({self.hostname})") from err

        except requests.exceptions.ConnectTimeout as err:
            log.debug(err)
            raise GeneralError(f"Connection timeout ({self.hostname})") from err

        except requests.exceptions.ConnectionError as err:
            log.debug(err)
            error = "Unknown connection error"

            try:
                if "Name or service not known" in err.args[0].reason.args[0]:
                    error = "Name or service not known"

                elif "Connection refused" in err.args[0].reason.args[0]:
                    error = "Connection refused"

            except (IndexError, AttributeError, TypeError):
                pass

            raise GeneralError(f"{error} ({self.hostname})") from err

        except requests.exceptions.ReadTimeout as err:
            log.error(err)
            raise GeneralError(f"Read timeout ({self.hostname})") from err

        log.debug(
            'Request time: [bold green]%2dms[/] "%s %s"',
            ((time() - start) * 1000),
            method.upper(),
            full_url,
        )

        try:
            response.raise_for_status()

        except requests.exceptions.HTTPError as err:
            raise APIError(err) from err

        return response

    @staticmethod
    def get_vendor() -> str:
        """
        Returns the vendor of the product (which always is Fortinet).

        Returns:
            Always returns "Fortinet" (what a cool method ;-)
        """
        return "Fortinet"

    @abstractmethod
    def get_version(self) -> str:
        """
        Gets the version of the corresponding system(s)
        """



"""
FortiClient EMS Class
"""

import logging
import pickle
import re
from pathlib import Path
from typing import Any, Dict, Optional

import requests

from fotoobo.exceptions import APIError, GeneralWarning

from .fortinet import Fortinet

log = logging.getLogger("fotoobo")


class FortiClientEMS(Fortinet):
    """
    Represents one FortiClient EMS (digital twin)
    """

    ALLOWED_HTTP_METHODS = ["DELETE", "GET", "PATCH", "POST"]

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        cookie_path: str = "",
        **kwargs: Any,
    ) -> None:
        """
        Set some initial parameters.

        Args:
            hostname:    The hostname of the FortiClient EMS to connect to
            username:    Username
            password:    Password
            cookie_path: Path to write the cookie files (no cookie if empty)
            **kwargs:    See Fortinet class for available arguments
        """
        super().__init__(hostname, **kwargs)
        self.api_url = f"https://{self.hostname}:{self.https_port}/api/v1"
        self.cookie_path = cookie_path
        self.password: str = password
        self.username: str = username
        self.type: str = "forticlientems"

    def api(  # pylint: disable=too-many-arguments
        self,
        method: str,
        url: str = "",
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        payload: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
    ) -> requests.models.Response:
        """
        API request to a FortiClientEMS device.

        Args:
            method:  Request method from [get, post]
            url:     Rest API URL to request data from
            headers: Additional headers (if needed)
            params:  Dictionary with parameters (if needed)
            payload: JSON body for post requests (if needed)
            timeout: The requests read timeout

        Returns:
            Response from the request
        """
        if not headers:
            headers = self.session.headers  # type: ignore

        return super().api(
            method, url, payload=payload, params=params, timeout=timeout, headers=headers
        )

    def get_version(self) -> str:
        """
        Get the FortiClient EMS version.
        According to Fortinet support the FortiClient EMS version may be read out of the endpoint
        /system/consts/get?system_update_time=1. This endpoint is not (yet) documented. Let's hope
        they support it in case of any issue.

        Returns:
            FortiClient EMS version
        """
        try:
            response = self.api("get", "/system/consts/get?system_update_time=1")

        except APIError as err:
            log.warning("'%s' returned: '%s'", self.hostname, err.message)
            raise GeneralWarning(f"{self.hostname} returned: {err.message}") from err

        try:
            ems_version: str = response.json()["data"]["System"]["VERSION"]

        except KeyError as err:
            log.warning("Did not find any FortiClient EMS version number in response")
            raise GeneralWarning(
                "Did not find any FortiClient EMS version number in response"
            ) from err

        return ems_version

    def login(self) -> int:
        """
        Login to the FortiClientEMS.

        Returns:
            Status code from the FortiClient EMS logon
        """
        status = 401
        cookie = Path(self.cookie_path).expanduser() / f"{self.hostname}.cookie"
        csrf = Path(self.cookie_path).expanduser() / f"{self.hostname}.csrf"

        if self.cookie_path:
            log.debug("Searching cookie and csrf token in '%s'", cookie.parents[0])

            if cookie.is_file() and csrf.is_file():
                log.debug("Cookie and csrf token both exist")
                with cookie.open("rb") as cookie_file:
                    self.session.cookies.update(pickle.load(cookie_file))  # type: ignore

                self.session.headers["Referer"] = f"https://{self.hostname}"
                self.session.headers["X-CSRFToken"] = csrf.read_text()

                try:
                    response = self.api("get", "/system/serial_number")
                    if (
                        "retval" in response.json()["result"]
                        and int(response.json()["result"]["retval"]) == 1
                    ):
                        log.debug(
                            "Session with given cookie is valid (status: '%s')",
                            response.status_code,
                        )
                        status = response.status_code

                except APIError as err:
                    log.debug("Session with given cookie is invalid (status: '%s')", err.code)
                    status = err.code

            else:
                log.debug("No cookie or csrf token found for '%s'", self.hostname)

        if status == 401:
            log.debug("Login to '%s'", self.hostname)
            payload = {"name": self.username, "password": self.password}
            response = self.api("post", "/auth/signin", payload=payload)

            if response.status_code == 200:
                self.session.headers["Referer"] = f"https://{self.hostname}"
                if match := re.match(r"csrftoken=(\S+);", response.headers["Set-Cookie"]):
                    csrf_token = match.group(1)
                    self.session.headers["X-CSRFToken"] = csrf_token

                if self.cookie_path:
                    log.debug("Saving cookie for '%s'", self.hostname)
                    try:
                        with cookie.open("wb") as cookie_file:
                            pickle.dump(self.session.cookies, cookie_file)

                    except FileNotFoundError as exc:
                        log.debug(exc)
                        log.warning("Unable to save cookie file '%s'", str(cookie.resolve()))

                    log.debug("Saving csrf token for '%s'", self.hostname)
                    try:
                        csrf.write_text(csrf_token)

                    except NameError as exc:
                        log.debug(exc)
                        log.warning("Unable to save csrf token file '%s'", str(csrf.resolve()))

            status = response.status_code

        return status

    def logout(self) -> int:
        """
        Logout from FortiClient EMS.

        Returns:
            Status code from the FortiClient EMS logout
        """
        response = self.api("get", "/auth/signout")
        log.debug("Logged out from '%s' (status: '%s')", self.hostname, response.status_code)
        return response.status_code



"""
The FortiManager commands
"""

import logging
from pathlib import Path

import typer

from fotoobo.cli.fmg import get_commands as get
from fotoobo.helpers import cli_path
from fotoobo.helpers.config import config as fotoobo_config
from fotoobo.inventory import Inventory
from fotoobo.tools import fmg

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fmg subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command(no_args_is_help=True)
def assign(
    adoms: str = typer.Argument(
        ...,
        help="The ADOMs to assign the global policy/objects to. Use 'fotoobo fmg get adoms' to "
        "get a list of available ADOMs. Separate multiple ADOMs by comma (no spaces).",
        metavar="[adoms]",
        show_default=False,
    ),
    policy: str = typer.Argument(
        ...,
        help="The global policy to assign",
        metavar="[policy]",
        show_default=False,
    ),
    host: str = typer.Argument(
        "fmg",
        help="The FortiManager to access (must be defined in the inventory).",
        metavar="[host]",
    ),
    smtp_server: str = typer.Option(
        None,
        "--smtp",
        "-s",
        help="The smtp configuration from the inventory to send potential errors to.",
        metavar="[server]",
    ),
    timeout: int = typer.Option(
        60,
        "--timeout",
        "-t",
        help="The timeout to wait for the FortiManager task to finish.",
        metavar="[timeout]",
    ),
) -> None:
    """
    Assign a global policy to a specified ADOM or to a list of ADOMs.
    """
    inventory = Inventory(fotoobo_config.inventory_file)
    result = fmg.assign(adoms=adoms, policy=policy, host=host, timeout=timeout)

    if smtp_server:
        if smtp_server in inventory.assets:
            result.send_messages_as_mail(inventory.assets[smtp_server], "error")

        else:
            log.warning("SMTP server '%s' not in found in inventory.", smtp_server)


@app.command(no_args_is_help=True)
def post(
    file: Path = typer.Argument(
        ..., help="JSON file with payload(s).", show_default=False, metavar="[file]"
    ),
    adom: str = typer.Argument(
        ..., help="The ADOM to issue the set command(s).", metavar="[adom]", show_default=False
    ),
    host: str = typer.Argument(
        "fmg",
        help="The FortiManager to access (must be defined in the inventory).",
        metavar="[host]",
    ),
    smtp_server: str = typer.Option(
        None,
        "--smtp",
        "-s",
        help="The smtp configuration from the inventory to send potential errors to.",
        metavar="[server]",
    ),
) -> None:
    """
    POST any valid JSON request to the FortiManager.

    Configure the FortiManager with any valid API call(s) given within the JSON file.
    """
    inventory = Inventory(fotoobo_config.inventory_file)
    result = fmg.post(file=file, adom=adom, host=host)

    if smtp_server:
        if smtp_server in inventory.assets:
            result.send_messages_as_mail(inventory.assets[smtp_server], "error", command=True)

        else:
            log.warning("SMTP server '%s' not in found in inventory.", smtp_server)


app.add_typer(get.app, name="get", help="FortiManager get commands.")



"""The Hashicorp Vault helper module for approle

The official API documentation is here: https://developer.hashicorp.com/vault/api-docs

We intend to use direct https requests without the dependency to another module (like hvac). This
makes this module very independent.
"""

import logging
from pathlib import Path
from typing import Any, Optional, Union

import requests
import urllib3

from fotoobo.exceptions.exceptions import GeneralError

log = logging.getLogger("fotoobo")


class Client:  # pylint: disable=too-many-instance-attributes
    """The approle helper class for the Hashicorp Vault

    This vault client gives you methods for your login with approle and maintaining token and data
    requests.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        url: str,
        namespace: str,
        data_path: str,
        role_id: str,
        secret_id: str,
        ssl_verify: Union[bool, str] = True,
        token_file: Optional[str] = None,
        token_ttl_limit: int = 0,
    ) -> None:
        """Initialize the vault client

        Args:
            url:                The URL of your vault service (eg. https://vault.local:443)
            namespace:          The namespace of your vault
            data_path:          The path where the vault data for fotoobo is stored
            role_id:            The approle role_id
            secret_id:          The approle secret_id
            ssl_verify:         Enable/disable SSL certificate check
                When ssl_verify is enabled you have to install a trusted SSL certificate onto
                the device you wish to connect to. If you set ssl_verify to false it will also
                disable the warnings in urllib3. This prevents unwanted SSL warnings to be
                logged.
            token_file:         The file to store the access token to. If no file is given the token
                                is not loaded or stored to a file and every execution will issue a
                                new token.
            token_ttl_limit:    Set a token limit. If the ttl of a token falls below this limit we
                                will automatically issue a new token. Default: 0
        """
        log.debug("Initialize the vault client")
        self.namespace: str = namespace
        self.data_path: str = data_path
        self.role_id: str = role_id
        self.secret_id: str = secret_id
        self.token: str = ""
        self.token_file: Optional[Path] = None
        self.token_ttl_limit: int = token_ttl_limit
        self.url: str = url.strip("/")
        self.ssl_verify: Union[bool, str] = ssl_verify
        if not self.ssl_verify:
            log.warning("SSL verify for vault service is disabled which may be a security issue")
            urllib3.disable_warnings(category=urllib3.exceptions.InsecureRequestWarning)

        log.debug("vault_client_url: '%s'", self.url)
        log.debug("vault_client_ssl_verify: '%s'", self.ssl_verify)
        log.debug("vault_client_namespace: '%s'", self.namespace)
        log.debug("vault_client_data_path: '%s'", self.data_path)
        log.debug("vault_client_role_id: '%s...%s'", self.role_id[:4], self.role_id[-4:])
        log.debug("vault_client_secret_id: '%s...%s'", self.secret_id[:4], self.secret_id[-4:])
        log.debug("vault_client_token_ttl_limit: '%s'", self.token_ttl_limit)

        if token_file:
            self.token_file = Path(token_file).expanduser()
            log.debug("vault_client_token_file: '%s'", self.token_file)
            self.load_token()

    def get_data(self, timeout: int = 3) -> Any:
        """Get data from the key/value store

        Args:
            timeout: The time before a request to the vault is cancelled

        Returns:
            The date from the vault or the error that occurred

        Raises:
            GeneralError: If no token could be retrieved
        """
        if not self.token and not self.get_token():
            error_message = "Unable to get vault token"
            log.error(error_message)
            raise GeneralError(error_message)

        url = f"{self.url}/{self.data_path.strip('/')}"
        log.debug("Get data from '%s'", url)
        headers = {
            "X-Vault-Token": self.token,
            "X-Vault-Namespace": self.namespace,
        }
        response = requests.get(url=url, headers=headers, timeout=timeout, verify=self.ssl_verify)

        if response.ok:
            log.debug("Response status_code is '%s'", response.status_code)
            return {"ok": response.json()}

        log.warning("Response is '%s %s'", response.status_code, response.reason)
        return {
            "error": {
                "status_code": response.status_code,
                "reason": response.reason,
                "content": response.content,
            }
        }

    def get_token(self, timeout: int = 3) -> bool:
        """Get a new token from the vault service by providing role_id and secret_id

        Args:
            timeout: The time before a request to the vault is cancelled
        """
        url = f"{self.url}/v1/auth/approle/login"
        log.debug("Get new token from '%s'", url)
        data = {"role_id": self.role_id, "secret_id": self.secret_id}
        self.token = ""

        try:
            response = requests.post(url, data=data, timeout=timeout, verify=self.ssl_verify)
            log.debug("Response status_code is '%s'", response.status_code)

            if response.ok:
                self.token = response.json()["auth"]["client_token"]
                if self.token_file:
                    self.save_token()

        except (requests.exceptions.SSLError, requests.exceptions.ConnectionError) as err:
            log.error("Request Error: %s", str(err))

        return bool(self.token)

    def load_token(self) -> bool:
        """Load the token from a file

        Open a text file with a vault token and validate the token against the vault.

        Returns:
            True if the token is valid, otherwise False
        """
        try:
            log.debug("Load token from file '%s'", str(self.token_file))
            self.token = self.token_file.read_text(encoding="utf-8")  # type: ignore
            self.validate_token()

        except FileNotFoundError:
            log.warning("Token file '%s' not found", str(self.token_file))

        return bool(self.token)

    def save_token(self) -> bool:
        """Save the vault token to a text file

        Returns:
            True if a token is set, otherwise False
        """
        if self.token:
            log.debug("Save token to file '%s'", str(self.token_file))
            self.token_file.write_text(self.token, encoding="utf-8")  # type: ignore

        return bool(self.token)

    def validate_token(self, timeout: int = 3) -> bool:
        """Check if the token still is valid

        Validate the Vault token against the Vault service. If the token ttl is lower then the
        token_ttl_limit the token is cleared. This prevents a token with ttl short before 0 to be
        used in future calls. Instead a new token should be requested.

        Args:
            timeout: The time before a request to the Vault service is cancelled

        Returns:
            True if the token is valid, otherwise False
        """
        url = f"{self.url}/v1/auth/token/lookup-self"
        log.debug("Check if vault token still is valid")
        headers = {"X-Vault-Token": self.token}

        try:
            response = requests.get(
                url=url, headers=headers, timeout=timeout, verify=self.ssl_verify
            )
            log.debug("Response status_code is '%s'", response.status_code)

            if response.ok:
                log.debug("Vault token is valid for '%s' seconds", response.json()["data"]["ttl"])
                log.debug("vault_client_token: '%s...%s'", self.token[:8], self.token[-4:])
                if response.json()["data"]["ttl"] < self.token_ttl_limit:
                    log.debug("Invalidate token due to ttl limit")
                    self.token = ""

            else:
                log.debug("Token is not valid")
                self.token = ""

        except (requests.exceptions.SSLError, requests.exceptions.ConnectionError) as err:
            self.token = ""
            log.error("Request Error: %s", str(err))

        return bool(self.token)



"""
The fotoobo convert commands
"""

import logging
from pathlib import Path

import typer

from fotoobo.helpers import cli_path
from fotoobo.helpers.files import load_json_file, save_json_file
from fotoobo.tools import convert

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fotoobo convert command callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command(no_args_is_help=True)
def checkpoint(
    infile: Path = typer.Argument(
        ...,
        help="The json file to read the Checkpoint objects from.",
        show_default=False,
        metavar="[infile]",
    ),
    outfile: Path = typer.Argument(
        ...,
        help="The json file to write the converted objects to.",
        show_default=False,
        metavar="[outfile]",
    ),
    obj_type: str = typer.Argument(
        ...,
        help="The type of objects to convert.",
        show_default=False,
        metavar="[type]",
    ),
    cache_dir: Path = typer.Argument(
        None,
        help="The cache directory to use.",
        show_default=False,
        metavar="[cache_dir]",
    ),
) -> None:
    """
    Convert Checkpoint assets into Fortinet objects.

    The Checkpoint objects have to be prepared in a json file. See
    https://fotoobo.readthedocs.io/en/latest/usage/convert.html for more information.
    """
    checkpoint_assets = load_json_file(infile)
    result = convert.checkpoint(checkpoint_assets, obj_type, outfile.name, cache_dir)
    save_json_file(outfile, result.get_result("fortinet_assets"))



"""
The tools module

Use this module to add business logic for your Fortinet infrastructure. This module uses the
relevant fotoobo and Fortinet modules. By default the tools functions are invoked by the cli app,
but they may also be accessed directly.
"""

from . import convert, ems, faz, fgt, fmg, get
from .greet import greet

__all__ = [
    "convert",
    "ems",
    "faz",
    "fgt",
    "fmg",
    "get",
    "greet",
]



"""
FortiGate configuration checker
"""

import logging
from typing import Any, Dict, List

from fotoobo.exceptions import GeneralError
from fotoobo.fortinet.fortigate_config import FortiGateConfig
from fotoobo.helpers.result import Result

log = logging.getLogger("fotoobo")


class FortiGateConfigCheck:
    """The FortiGate configuration check class"""

    def __init__(self, config: FortiGateConfig, checks: Any, result: Result[Any]) -> None:
        """
        Initialize the configuration checker.

        Args:
            config: The FortiGate configuration
            checks: The checks to do against the FortiGate configuration
        """
        self.allowed_checks: List[str] = ["count", "exist", "value", "value_in_list"]
        self.config = config
        self.checks = checks
        self.result = result

    def add_message(self, chk: Dict[str, Any], msg: str) -> None:
        """
        Generates a styled message and appends it to the results.

        Args:
            chk: The check which generated the message
            msg: The message to send to the messages list
        """
        check_name = f" (check_name: [var]{chk['name']}[/])" if "name" in chk else ""
        message = f"[chk]{chk['type']}[/]: {msg}{check_name}"
        log.info(message)
        self.result.push_message(self.config.info.hostname, message)

    def execute_checks(self) -> Result[Any]:  # pylint: disable=too-many-branches
        """
        Execute the FortiGate configuration checks.

        After initializing a FortiGateConfigCheck object you can run this method to actually run
        the checks and write the results into the results object.
        """
        if not self.checks:
            log.error("There are no checks defined")
            raise GeneralError("There are no checks defined")

        for check in self.checks:
            # applying info filter
            skip = False
            for info_filter in check.get("filter-info", []):
                if str(check["filter-info"]).startswith("<"):
                    if not getattr(self.config.info, info_filter) < str(
                        check["filter-info"][info_filter][1:]
                    ):
                        log.debug("Skipping check due to filter-info '%s'", info_filter)
                        skip = True

                elif str(check["filter-info"]).startswith(">"):
                    if not getattr(self.config.info, info_filter) > str(
                        check["filter-info"][info_filter][1:]
                    ):
                        log.debug("Skipping check due to filter-info '%s'", info_filter)
                        skip = True

                else:
                    if not getattr(self.config.info, info_filter) == str(
                        check["filter-info"][info_filter]
                    ):
                        log.debug("Skipping check due to filter-info '%s'", info_filter)
                        skip = True

            # applying config_filter
            for conf_filter in check.get("filter-config", []):
                config = self.config.get_configuration(check["scope"], conf_filter)
                if not config == check["filter-config"][conf_filter]:
                    log.debug("Skipping check due to filter-config '%s'", conf_filter)
                    skip = True

            if skip:
                continue

            # check if needed check keys are present
            if miss := ("type", "scope", "path", "checks") - check.keys():
                log.error("Key(s) '%s' missing in '%s'", miss, check.get("name", "unnamed check"))
                continue

            # check if checks are defined
            if not check["checks"]:
                log.error("No checks defined in '%s'", check.get("name", "unnamed check"))
                continue

            if not check["type"] in self.allowed_checks:
                log.error(
                    "Check type '%s' not available in '%s'",
                    check.get("type"),
                    check.get("name", "unnamed check"),
                )
                continue

            if check["scope"] == "global":
                config = self.config.get_configuration(check["scope"], check["path"])
                getattr(self, "_check_" + check["type"])(config, check)

            if check["scope"] == "vdom":
                if self.config.info.vdom == "0":
                    if check["path"].startswith("/system/"):
                        config = self.config.get_configuration("global", check["path"])

                    else:
                        config = self.config.get_configuration("vdom", "/root" + check["path"])

                    getattr(self, "_check_" + check["type"])(config, check)

                elif self.config.info.vdom == "1":
                    for vdom in self.config.get_vdoms():
                        config = self.config.get_configuration(
                            check["scope"], vdom + "/" + check["path"]
                        )
                        getattr(self, "_check_" + check["type"])(config, check)

        return self.result

    def _check_count(self, config: Any, chk: Dict[str, Any]) -> None:
        """
        Check the configuration list count.

        Args:
            config: FortiGate configuration part to check
            chk:    The check dict to process
        """
        if isinstance(config, list):
            conf_len = len(config)
            for key, value in chk["checks"].items():
                # pylint: disable=too-many-boolean-expressions
                if (
                    (key == "eq" and not conf_len == int(value))
                    or (key == "gt" and not conf_len > int(value))
                    or (key == "lt" and not conf_len < int(value))
                ):
                    self.add_message(
                        chk,
                        f"count of [var]{chk['path']}[/] is not [var]{key}[/] [var]{value}[/]",
                    )

        else:
            log.warning("'%s' is not a configuration list", chk["path"])

    def _check_exist(self, config: Any, chk: Dict[str, Any]) -> None:
        """
        Check if a configuration option is present (or not) regardless of its value.

        Args:
            config: FortiGate configuration part to check
            chk:    The check dict to process
        """
        for key, value in chk["checks"].items():
            if bool(key in config) != value:
                self.add_message(
                    chk,
                    f"key [var]{key}[/] in [var]{chk['path']}[/] is not [var]{value}[/]",
                )

    def _check_value(self, config: Any, chk: Dict[str, Any]) -> None:
        """
        Do the checks for a configuration value. It checks if the configuration option is present
        and if the value is set as given in the check bundle.

        Args:
            config: FortiGate configuration part to check
            chk:    The check dict to process
        """
        for key, value in chk["checks"].items():
            msg_key = f"[var]{key}[/]"
            msg_path = f"[var]{chk['path']}[/]"

            if key in config:
                log.debug("Key '%s' in '%s' is '%s'", msg_key, msg_path, config[key])
                if str(value) != config[key]:
                    self.add_message(
                        chk,
                        f"key {msg_key} in {msg_path} is not [var]{value}[/]",
                    )

            else:
                if not chk.get("ignore_missing", False):
                    self.add_message(chk, f"key {msg_key} does not exist in config")

    def _check_value_in_list(self, config: Any, chk: Dict[str, Any]) -> None:
        """
        Do the checks for set configuration. It checks if the configuration option is present in a
        and configuration list and if the value is set as given in the check bundle.

        Args:
            config: FortiGate configuration part to check
            chk:    The check dict to process
        """
        inverse: bool = chk.get("inverse", False)
        msg_not = "" if inverse else "not "

        for key, val in chk["checks"].items():
            msg_key = f"[var]{key}[/]"
            msg_val = f"[var]{val}[/]"
            msg_path = f"[var]{chk['path']}[/]"
            exist = False

            for conf in config:
                if key in conf and val == conf[key]:
                    exist = True

            if not exist ^ inverse:
                self.add_message(
                    chk,
                    f"{msg_key}: {msg_val} {msg_not}in {msg_path}",
                )



"""
The FortiGate get commands
"""

# pylint: disable=anomalous-backslash-in-string
import logging

import typer

from fotoobo.helpers import cli_path
from fotoobo.tools import fgt

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fgt get subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def version(
    host: str = typer.Argument(
        "",
        help="The FortiGate hostname to access (must be defined in the inventory). "
        "\[default: <all>]",
        show_default=False,
        metavar="[host]",
    )
) -> None:
    """
    Get the FortiGate(s) version(s).

    The optional argument [host] makes this command somewhat magic. If you omit \[host] it searches
    for all devices of type 'fortigate' in the inventory and tries to get their FortiOS version.
    """
    result = fgt.get.version(host)
    result.print_result_as_table(title="FortiGate Versions", headers=["FortiGate", "Version"])



"""
The FortiAnalyzer get commands
"""

import logging

import typer

from fotoobo.helpers import cli_path
from fotoobo.tools import faz

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The faz get subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def version(
    host: str = typer.Argument(
        "faz",
        help="The FortiAnalyzer hostname to access (must be defined in the inventory).",
        metavar="[host]",
    )
) -> None:
    """
    Get the FortiAnalyzer version.
    """
    result = faz.get.version(host)
    result.print_result_as_table(
        title="FortiAnalyzer Version",
        headers=["FortiAnalyzer", "Version"],
    )



"""
The FortiClient EMS monitor commands
"""

import logging
from pathlib import Path
from typing import Optional, Union

import typer

from fotoobo.helpers import cli_path
from fotoobo.helpers.files import save_json_file
from fotoobo.tools.ems import monitor

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


HELP_TEXT_ARGUMENT_EMS = (
    "The FortiClient EMS hostname to access (must be defined in the inventory)."
)
HELP_TEXT_OPTION_OUTPUT_FILE = "The file to write the output to."
HELP_TEXT_OPTION_TEMPLATE = "The jinja2 template to use (use with -o)."
HELP_TEXT_TEMPLATE = (
    "If you add a template with the -t option you may render the output with any Jinja2 template "
    "file. You may use any of the given data returned from the FortiClient EMS. Additionally there "
    "are enriched variables under 'fotoobo' which you may also use in your template."
)


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The ems get subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command(help="Monitor the FortiClient EMS connections.\n\n" + HELP_TEXT_TEMPLATE)
def connections(
    host: str = typer.Argument(
        "ems",
        help=HELP_TEXT_ARGUMENT_EMS,
        metavar="[host]",
    ),
    output_file: Union[None, Path] = typer.Option(
        None,
        "--output",
        "-o",
        help=HELP_TEXT_OPTION_OUTPUT_FILE,
        metavar="[output]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
    template_file: Optional[Path] = typer.Option(
        None,
        "--template",
        "-t",
        help=HELP_TEXT_OPTION_TEMPLATE,
        metavar="[template]",
    ),
) -> None:
    """
    Monitor the FortiClient EMS connections.
    """
    result = monitor.connections(host)
    data = result.get_result(host)

    if output_file:
        log.debug("output_file is: '%s'", output_file)

        if template_file:
            log.debug("template_file is: '%s'", template_file)
            result.save_with_template(host, template_file, output_file)

        else:
            # write to file without a template (raw output)
            save_json_file(output_file, data)

    else:
        # if no output file is given just pretty print the output to the console
        if raw:
            result.print_raw()

        else:
            conns = []
            for connection in result.get_result(host)["data"]:
                conns.append(
                    {
                        "token": connection["token"]
                        .replace("offline", "Offline ")
                        .replace("online", "Online"),
                        "value": connection["value"],
                        "name": connection["name"],
                    }
                )
            result.print_table_raw(
                conns, ["Token", "Count", "Description"], title="FortiClient connection status"
            )


@app.command(
    help="Monitor the endpoint management status in FortiClient EMS.\n\n" + HELP_TEXT_TEMPLATE
)
def endpoint_management_status(
    host: str = typer.Argument(
        "ems",
        help=HELP_TEXT_ARGUMENT_EMS,
        metavar="[host]",
    ),
    output_file: Union[None, Path] = typer.Option(
        None,
        "--output",
        "-o",
        help=HELP_TEXT_OPTION_OUTPUT_FILE,
        metavar="[output]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
    template_file: Optional[Path] = typer.Option(
        None,
        "--template",
        "-t",
        help=HELP_TEXT_OPTION_TEMPLATE,
        metavar="[template]",
    ),
) -> None:
    """
    Monitor the endpoint management status in FortiClient EMS.
    """
    result = monitor.endpoint_management_status(host)
    data = result.get_result(host)

    if output_file:
        log.debug("output_file is: '%s'", output_file)

        if template_file:
            result.save_with_template(host, template_file, output_file)

        else:
            # write to file without a template (raw output)
            save_json_file(output_file, data)

    else:
        # if no output file is given just pretty print the output to the console
        if raw:
            result.print_raw()

        else:
            endpoints = []
            for endpoint in result.get_result(host)["data"]:
                endpoints.append(
                    {
                        "name": endpoint["name"],
                        "value": endpoint["value"],
                    }
                )
            result.print_table_raw(
                endpoints, ["Status", "Count"], title="FortiClient management status"
            )


@app.command(help="Monitor the endpoint OS versions in FortiClient EMS.\n\n" + HELP_TEXT_TEMPLATE)
def endpoint_os_versions(
    host: str = typer.Argument(
        "ems",
        help=HELP_TEXT_ARGUMENT_EMS,
        metavar="[host]",
    ),
    output_file: Union[None, Path] = typer.Option(
        None,
        "--output",
        "-o",
        help=HELP_TEXT_OPTION_OUTPUT_FILE,
        metavar="[output]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
    template_file: Optional[Path] = typer.Option(
        None,
        "--template",
        "-t",
        help=HELP_TEXT_OPTION_TEMPLATE,
        metavar="[template]",
    ),
) -> None:
    """
    Monitor the endpoint os versions in FortiClient EMS.
    """
    result = monitor.endpoint_os_versions(host)
    data = result.get_result(host)

    if output_file:
        log.debug("output_file is: '%s'", output_file)

        if template_file:
            result.save_with_template(host, template_file, output_file)

        else:
            # write to file without a template (raw output)
            save_json_file(output_file, data)

    else:
        # if no output file is given just pretty print the output to the console
        if raw:
            result.print_raw()

        else:
            versions = []
            for _os, os_versions in result.get_result(host)["data"].items():
                for line in os_versions:
                    versions.append(
                        {"os": _os[10:], "fc_version": line["name"], "count": line["value"]}
                    )
            result.print_table_raw(versions, ["OS", "FC Version", "Count"])

            totals = []
            for _os, count in result.get_result(host)["fotoobo"].items():
                totals.append({"os": _os[10:], "count": count})

            result.print_table_raw(totals, ["OS", "Total"], title="Summary")


@app.command(
    help="Get the amount of FortiClient EMS devices which are online but policy is not in sync.\n\n"
    + HELP_TEXT_TEMPLATE
)
def endpoint_outofsync(
    host: str = typer.Argument(
        "ems",
        help=HELP_TEXT_ARGUMENT_EMS,
        metavar="[host]",
    ),
    output_file: Union[None, Path] = typer.Option(
        None,
        "--output",
        "-o",
        help=HELP_TEXT_OPTION_OUTPUT_FILE,
        metavar="[output]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
    template_file: Optional[Path] = typer.Option(
        None,
        "--template",
        "-t",
        help=HELP_TEXT_OPTION_TEMPLATE,
        metavar="[template]",
    ),
) -> None:
    """
    Get amount of FortiClient EMS devices which are online but policy not in sync.
    """
    result = monitor.endpoint_online_outofsync(host)
    data = result.get_result(host)

    if output_file:
        log.debug("output_file is: '%s'", output_file)

        if template_file:
            result.save_with_template(host, template_file, output_file)

        else:
            # write to file without a template (raw output)
            save_json_file(output_file, data)

    else:
        # if no output file is given just pretty print the output to the console
        if raw:
            result.print_raw()

        else:
            result.print_table_raw(
                [{"name": "out of sync", "count": result.get_result(host)["fotoobo"]["outofsync"]}],
                [],
                title="FortiClients not in synch",
            )


@app.command(help="Monitor the FortiClient EMS license.\n\n" + HELP_TEXT_TEMPLATE)
def license(  # pylint: disable=redefined-builtin
    host: str = typer.Argument(
        "ems",
        help=HELP_TEXT_ARGUMENT_EMS,
        metavar="[host]",
    ),
    output_file: Union[None, Path] = typer.Option(
        None,
        "--output",
        "-o",
        help=HELP_TEXT_OPTION_OUTPUT_FILE,
        metavar="[output]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
    template_file: Optional[Path] = typer.Option(
        None,
        "--template",
        "-t",
        help=HELP_TEXT_OPTION_TEMPLATE,
        metavar="[template]",
    ),
) -> None:
    """
    Monitor the FortiClient EMS license.
    """
    result = monitor.license(host)
    data = result.get_result(host)

    if output_file:
        log.debug("output_file is: '%s'", output_file)

        if template_file:
            result.save_with_template(host, template_file, output_file)

        else:
            # write to file without a template (raw output)
            save_json_file(output_file, data)

    else:
        # if no output file is given just pretty print the output to the console
        if raw:
            result.print_raw()

        else:
            licenses = []
            for key, value in result.get_result(host)["data"].items():
                licenses.append({"key": key, "value": value})

            result.print_table_raw(
                licenses, ["Key", "Value"], title="FortiClient EMS license information"
            )

            licenses = []
            for key, value in result.get_result(host)["fotoobo"].items():
                licenses.append({"key": key, "value": value})

            result.print_table_raw(
                licenses, ["Key", "Value"], title="FortiClient EMS license summary"
            )


@app.command()
def system(
    host: str = typer.Argument(
        "ems",
        help=HELP_TEXT_ARGUMENT_EMS,
        metavar="[host]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
) -> None:
    """
    Monitor the FortiClient EMS system information.
    """
    result = monitor.system(host)
    data = result.get_result(host)

    if raw:
        result.print_raw()

    else:
        _ = data.pop("license", {})  # pop "license" key as ist is not used (it's another command)
        result.print_table_raw(
            [
                {"key": "hostname", "hostname": data["name"]},
                {"key": "system_time", "system_time": data["system_time"]},
            ],
            ["Key", "Value"],
            title="FortiClient EMS status",
        )



"""
The fotoobo Result class
"""

import json
import re
import smtplib
from pathlib import Path
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

import jinja2
from rich.console import Console
from rich.pretty import pprint
from rich.table import Table
from rich.theme import Theme

from fotoobo.exceptions import GeneralWarning
from fotoobo.helpers import cli_path

ftb_theme = Theme({"var": "white", "ftb": "#FF33BB bold", "chk": "green"})

T = TypeVar("T")


class Result(Generic[T]):
    """
    This class represents a Result of an operation in fotoobo.

    This dataset is meant to be the generic result structure for any tool inside fotoobo.
    It can then be rendered to some command line output (CLI) or JSON response (REST API).
    """

    def __init__(self) -> None:
        """
        Create the FotooboResult object
        """
        # The devices where the processing has been successful
        self.successful: List[str] = []

        # The devices where the processing gave an error
        self.failed: List[str] = []

        # The total number of devices processed
        self.total: int = 0

        # Add messages for each device
        self.messages: Dict[str, List[Dict[str, str]]] = {}

        # The results for each device
        self.results: Dict[str, T] = {}

        # Console object for rich output
        self.console = Console(theme=ftb_theme)

    def push_result(self, key: str, data: T, successful: bool = True) -> None:
        """
        Add a result for the given key

        Args:
            key:        The key to push the results for
            data:       The output data for this key
            successful: Whether the call has been successful or not [default: True]
        """
        self.results[key] = data
        if successful:
            self.successful.append(key)

        else:
            self.failed.append(key)

    def push_message(self, host: str, message: str, level: str = "info") -> None:
        """
        Add a message for the host

        Args:
            host:    The host to add the message for
            message: The message to add
            level:   The level to assign to this message, used for later filtering
                     (use for example "info", "warning", "error")
        """
        if host not in self.messages:
            self.messages[host] = []

        self.messages[host].append({"message": message, "level": level})

    def get_messages(self, host: str) -> List[Dict[str, str]]:
        """
        Return all the messages for the host given

        Args:
            host: The host to get the messages for

        Returns:
            A list of the messages pushed for this host. If there are no messages for this host, it
            will return an empty list.
        """

        if host not in self.messages:
            return []

        return self.messages[host]

    def print_messages(self, only_host: Union[str, None] = None) -> None:
        """
        Print the messages to the console

        Args:
            only_host:  Only print the messages for host only_host. If None is given, print all
                        messages.
        """
        out_messages = []

        for host, _messages in self.messages.items():
            if only_host and host != only_host:
                continue

            out_messages += [f"[hst]{host}[/]: {message['message']}" for message in _messages]

        for message in out_messages:
            self.console.print(message)

    def get_result(self, host: str) -> T:
        """
        Return the result pushed by this

        Args:
            host: The name of the host to retrieve the result for

        Returns:
            The results stored before with push_results, or None if the host does not exist

        Raises:
            GeneralWarning: If there is no result for host
        """
        if host not in self.results:
            raise GeneralWarning(f"Host {host} is not in results.")

        return self.results[host]

    def all_results(self) -> Dict[str, T]:
        """
        Return all results

        Returns:
            The results as a dictionary of the following form:
                { '<name of the host>': <data> }
            where <data> may be of any type
        """
        return self.results

    def print_result_as_table(
        self,
        title: str = "",
        auto_header: bool = False,
        headers: Optional[List[str]] = None,
    ) -> None:
        """
        Print a table from given data as list or dict.

        Args:
            title:       Set the preferred title for the table
            auto_header: Whether to show the headers (default: off)
            headers:     Set the headers (if needed)

        Raises:
            GeneralWarning: If the data cannot be interpreted as a table
        """
        if not headers:
            headers = []

        data: List[Dict[str, Any]] = []
        for host, result in self.results.items():
            if isinstance(result, dict):
                data.append({"key": host, **result})

            else:
                data.append({"key": host, "value": result})

        self.print_table_raw(data, headers, auto_header, title)

    def print_table_raw(
        self,
        data: List[Dict[str, Any]],
        headers: Optional[List[str]] = None,
        auto_header: bool = False,
        title: str = "",
    ) -> None:
        """
        Print the data given as a rich table to the console

        Args:
            data:        The data to print formatted as rich.table.Table will expect it
            headers:     The headers for the table
            auto_header: Whether to show the headers or not
            title:       The title for the table
        """
        if not isinstance(data, list):
            raise GeneralWarning("data for print_table_raw must be a list of dicts.")

        table = Table(title=title, show_header=auto_header or bool(headers))
        if auto_header:
            for heading in data[0].keys():
                table.add_column(heading)

        elif headers:
            for heading in headers:
                table.add_column(heading)

        for line in data:
            _values = line.values()

            # If an item in line is a dict or list we should pretty print it
            values: List[str] = []

            for entry in _values:
                if isinstance(entry, (dict, list)):
                    values.append(json.dumps(entry, indent=4))

                else:
                    values.append(str(entry))

            table.add_row(*values)

        self.console.print(table)

    def print_raw(self, key: Optional[str] = None) -> None:
        """
        Print the raw data from Result() in pretty format.

        Args:
            key: Print only the result for the host given
                 (default: print all results)
        """
        if key:
            data = {key: self.get_result(key)}

        else:
            data = {}
            for host in self.all_results():
                data[host] = self.results[host]

        pprint(data, expand_all=True)

    def save_with_template(self, host: str, template_file: Path, output_file: Path) -> None:
        """
        Saves a data structure to a file with a given Jinja2 template. The data structure and the
        variables you can use in the template file depend on the utility the data comes from. See
        the docs of the used utility to see what variables you're intended to use.

        Args:
            data:          The data used in the template
            template_file: Filename of the Jinja2 template file
            output_file:   The file to write the output to
        """
        template_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_file.parent), trim_blocks=True, autoescape=True
        )
        template = template_env.get_template(template_file.name)
        output = template.render(self.results[host])
        output_file.write_text(output, encoding="UTF-8")

    def send_messages_as_mail(
        self,
        smtp_server: Any,
        levels: Union[List[str], str, None] = None,
        count: bool = False,
        command: bool = False,
    ) -> None:
        """
        Send an e-mail with the messages collected up until the call

        Args:
            command:        Whether to write the issued command into the message (default: false)
            count:          Whether to write the message count into the message (default: false)
            smtp_server:    The smtp server from inventory to use
            levels:         The levels to output:
                            - None means all messages will be output (default)
                            - 'level' means only messages with level='level' will be output
                            - ['level1', 'level2'] like 2nd option, but all levels given will get
                            output
        """
        out_messages: List[str] = []
        for host, messages in self.messages.items():
            for message in messages:
                if not levels or message["level"] in levels:
                    msg = re.sub(r"\[.*?\]", "", message["message"])  # remove rich formatting
                    out_messages.append(f"{host}: {msg}")

        if out_messages:
            body = "To:" + smtp_server.recipient + "\n"
            body += "From:" + smtp_server.sender + "\n"
            body += "Subject:" + smtp_server.subject + "\n\n"

            for message_line in out_messages:
                body += message_line + "\n"

            if command and cli_path:
                body += "\ncommand: " + " ".join(cli_path) + "\n"

            if count:
                adds = "s" if len(out_messages) > 1 else ""
                body += "\n" + str(len(out_messages)) + " message" + adds + " in list"

            # Prepare server connection and send mail
            smtp_server.port = getattr(smtp_server, "port", 25)
            with smtplib.SMTP(smtp_server.hostname, smtp_server.port) as mail_server:
                # server.set_debuglevel(1)
                mail_server.sendmail(smtp_server.sender, smtp_server.recipient, body)



"""
The beautiful output helper
"""

import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

from rich.console import Console


def print_logo() -> None:
    """
    Print the beautiful fotoobo logo
    """
    logo_console = Console(style="#FF33BB bold")
    logo_console.print("╭───┐┌───┐┌───╮")
    logo_console.print("└───┘└───┘└───┘")
    logo_console.print(" f o t o o b o ")
    logo_console.print(" ━━━━━━━━━━━━━ ")
    logo_console.print(" make IT easy  ")
    logo_console.print("┌───┐┌───┐┌───┐")
    logo_console.print("╰───┘└───┘└───╯")


def write_policy_to_html(data: List[Dict[str, Any]], out_file: Path) -> None:
    """
    Write a Firewall policy to a HTML file

    Args:
        data:     List of Dicts with data
        out_file: Filename to write the HTML output to
    """
    ignored_rows = ["status", "global-label", "send-deny-packet"]
    table = '<table class="table table-bordered">\n'
    cols = len(data[0]) - len(ignored_rows)

    # Create the table's column headers
    table += "<thead><tr>\n"
    headers = list(data[0])
    for header in headers:
        if header not in ignored_rows:
            table += f"<th>{header}</th>\n"

    table += '</tr><thead><tbody id="myTable">\n'

    # Create the table's row data
    label = ""
    for line in data:
        if "global-label" in line and line["global-label"] and line["global-label"] != label:
            table += f'<tr><th colspan="{cols}" style="text-align: center;">'
            table += line["global-label"]
            table += f'<span style="display: none">{line["groups"]}</span>'
            table += "</th></tr>\n"
            label = line["global-label"]

        style = ""
        if "_hitcount" in line:
            style = ' style="background-color:#EEE"' if int(line["_hitcount"]) == 0 else ""

        if "status" in line:
            style = ' style="background-color:#FCC"' if int(line["status"]) == 0 else style

        table += f"<tr{style}>"
        for key, value in line.items():
            if key not in ignored_rows:
                if key == "_last_hit":
                    value = datetime.fromtimestamp(value).strftime("%d.%m.%Y") if value > 0 else 0

                if key == "action":
                    value = ["deny", "accept", "ipsec", "ssl-vpn", "redirect", "isolate"][value]
                    if value == "deny" and "send-deny-packet" in line:
                        value += f" ({['drop', 'reset'][line['send-deny-packet']]})"

                value = "<br />".join(value) if isinstance(value, list) else str(value)
                table += f"<td>{value}</td>"

        table += "</tr>\n"

    table += "</tbody></table>"

    html_header = """
        <!DOCTYPE html>
        <html lang="de">
        <head>
        <title>Firewall-Policies</title>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <meta name="description" content="script from fotoobo">
        <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css" rel="stylesheet">
        <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.6.0/jquery.min.js"></script>
        <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.bundle.min.js"></script>
        <style type="text/css">
            @media print {
                @page { size: A4 landscape; }
                body { zoom: 80%;}
                .noprint { display: none; }
            }
            body { margin-bottom: 10px; }
        </style>
        </head>
        <body>
        <div class="container-fluid mt-3">
        <h2>Firewall-Policies</h2>
        <div class="noprint">Type something in the input field to filter the table</div>
        <div class="noprint">
            <input class="form-control" id="myInput" type="text" placeholder="Search..">
            <br />
        </div>
        <div>
            <!--<span style="color:#EEE">&#9632;</span> : nohits | -->
            <span style="color:#FCC">&#9632;</span> : disabled
        </div>
        """

    html_footer = """
        <script>
        $(document).ready(function(){
            $("#myInput").on("keyup", function() {
                var value = $(this).val().toLowerCase();
                $("#myTable tr").filter(function() {
                $(this).toggle($(this).text().toLowerCase().indexOf(value) > -1)
                });
            });
        });
        </script>
        """
    now = datetime.now().strftime("%d.%m.%Y %H:%M")
    html_footer += '<div class="container-fluid text-center">'
    html_footer += f"created {now} on {os.uname().nodename} ({os.uname().sysname})"
    html_footer += "</div>"
    html_footer += """
    </body>
    </html>
    """

    out_file.write_text(html_header + table + html_footer, encoding="UTF-8")



"""
The FortiGate check commands
"""

import logging
from pathlib import Path
from typing import Dict, Optional

import typer
from rich.pretty import pprint

from fotoobo.helpers import cli_path
from fotoobo.helpers.config import config
from fotoobo.helpers.files import save_json_file
from fotoobo.helpers.result import Result
from fotoobo.inventory.inventory import Inventory
from fotoobo.tools import fgt

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


HELP_TEXT_OPTION_OUTPUT_FILE = "The file to write the output to."
HELP_TEXT_OPTION_TEMPLATE = "The jinja2 template to use (use with -o)."
HELP_TEXT_TEMPLATE = (
    "If you add a template with the -t option you may render the output with any Jinja2 template "
    "file. You may use any of the given data returned from the FortiClient EMS. Additionally there "
    "are enriched variables under 'fotoobo' which you may also use in your template."
)


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fgt check subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def hamaster(
    host: str = typer.Argument(
        "fmg",
        help="The FortiManager hostname to access (must be defined in the inventory).",
        metavar="[host]",
    ),
    output_file: Optional[Path] = typer.Option(
        None,
        "--output",
        "-o",
        help=HELP_TEXT_OPTION_OUTPUT_FILE,
        metavar="[output]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
    smtp_server: str = typer.Option(
        None,
        "--smtp",
        help="The smtp configuration from the inventory.",
        metavar="[server]",
        show_default=False,
    ),
    template_file: Optional[Path] = typer.Option(
        None,
        "--template",
        "-t",
        help=HELP_TEXT_OPTION_TEMPLATE,
        metavar="[template]",
    ),
) -> None:
    """
    Check the FortiGate HA master.

    Although this command checks the HA master status of FortiGates you have to specify a
    FortiManager to access. The command searches for all FortiGate clusters in the FortiManager
    and checks if the designated primary node really is the HA master node.

    The optional argument 'host' makes this command somewhat magic. If you omit 'host' it searches
    for all devices in the default FortiManager (fmg) in the inventory.
    """
    inventory = Inventory(config.inventory_file)
    result = fgt.monitor.hamaster(host)
    data = {"fotoobo": result.all_results()}

    if smtp_server:
        if smtp_server in inventory.assets:
            result.send_messages_as_mail(
                inventory.assets[smtp_server],
                ["warning", "error"],
                count=True,
                command=True,
            )

        else:
            log.warning("SMTP server '%s' not in found in inventory.", smtp_server)

    if output_file:
        log.debug("output_file is: '%s'", output_file)

        if template_file:
            log.debug("template_file is: '%s'", template_file)
            output: Result[Dict[str, Dict[str, str]]] = Result()
            output.push_result("_hamaster", data)
            output.save_with_template("_hamaster", template_file, output_file)

        else:
            # write to file without a template (raw output)
            save_json_file(output_file, data)

    else:
        # if no output file is given just pretty print the output to the console
        if raw:
            pprint(data, expand_all=True)

        else:
            result.print_result_as_table(
                headers=["FortiGate Cluster", "Status"],
                title="FortiGate HA master status",
            )



"""
The Fortinet module

Here we add all the functions, classes and methods for accessing and interacting with the Fortinet
infrastructure. This module is meant to be held very generic.
"""

from .fortianalyzer import FortiAnalyzer
from .forticlientems import FortiClientEMS
from .fortigate import FortiGate
from .fortimanager import FortiManager

__all__ = ["FortiAnalyzer", "FortiClientEMS", "FortiGate", "FortiManager"]



"""
Some helper functions for file manipulation.
"""

import json
import logging
import re
from ftplib import FTP, FTP_TLS
from pathlib import Path
from typing import Any, Dict, List, Union
from zipfile import ZIP_DEFLATED, ZipFile

import yaml

from fotoobo.exceptions import GeneralError

log = logging.getLogger("fotoobo")


def create_dir(directory: Path) -> None:
    """
    Try to create a given directory if it does not exist.

    Args:
        directory: The directory to create (in case it not already exists)

    Raises:
        OSError: Error with message
    """
    if not directory.is_dir():
        try:
            directory.mkdir(parents=True)

        except OSError as err:
            raise GeneralError(f"Unable to create directory {directory}") from err


def file_to_ftp(file: Path, server: Any) -> int:
    """
    Upload a file to a ftp server.

    Args:
        file:   The source file to upload
        server: The ftp sever definition dict containing the following keys:
            * hostname
            * directory
            * username
            * password
            * protocol (default="sftp")

    Returns:
        Return code
    """
    return_code = 0

    if file.is_file():
        if hasattr(server, "protocol"):
            protocol = server.protocol

        else:
            protocol = "sftp"

        if protocol == "sftp":
            with FTP_TLS(server.hostname) as ftp:
                log.debug("SFTP transfer for '%s'", server.hostname)
                ftp.sendcmd(f"USER {server.username}")
                ftp.sendcmd(f"PASS {server.password}")
                ftp.cwd(server.directory)

                with file.open("rb") as ftp_file:
                    response = ftp.storbinary(f"STOR {file.name}", ftp_file)
                    if response != "226 Transfer complete.":
                        if code := re.search(r"^([0-9]{0,3})\s", response):
                            return_code = int(code[1])

        elif protocol == "ftp":
            with FTP(server.hostname, server.username, server.password) as ftp:
                log.debug("FTP transfer for '%s'", server.hostname)
                ftp.cwd(server.directory)

                with file.open("rb") as ftp_file:
                    response = ftp.storbinary(f"STOR {file.name}", ftp_file)
                    if response != "226 Transfer complete.":
                        if code := re.search(r"^([0-9]{0,3})\s", response):
                            return_code = int(code[1])

        else:
            raise GeneralError(f'Unknown FTP protocol "{protocol}" for server "{server.hostname}"')

    else:
        return_code = 666

    return return_code


def file_to_zip(src: Path, dst: Path, level: int = 9) -> None:
    """
    Zip (compress) a file.

    Args:
        src:   The source file
        dst:   The destination file (zipfile). If the dst file ends with extension '.zip' the inner
               file will omit the '.zip' extension.
        level: Compression level, as accepted by `zipfile.ZipFile`

    Returns:
        Return code
    """
    if level < 0 or level > 9:
        raise GeneralError("zip level must between 0 and 9")

    inner_file = dst if not dst.suffix == ".zip" else Path(dst.name[0:-4])
    if src.is_file():
        with ZipFile(dst, "w", ZIP_DEFLATED, compresslevel=level) as archive:
            archive.write(src, arcname=inner_file.name)


def load_json_file(json_file: Path) -> Union[List[Any], Dict[str, Any], None]:
    """
    Loads the content of a json file into a list or dict.

    Args:
        json_file: The path to the json file to load

    Returns:
        List or dict with json data from filename or None if file is not found
    """
    content = None
    if json_file.is_file():
        with json_file.open(encoding="UTF-8") as in_file:
            content = json.load(in_file)

    return content


def load_yaml_file(yaml_file: Path) -> Union[List[Any], Dict[str, Any], None]:
    """
    Loads the content of a yaml file into a list or dict.

    Args:
        yaml_file: The file of the yaml file to load

    Returns:
        Yaml data from filename
    """
    content = None
    if yaml_file.is_file():
        content = yaml.safe_load(yaml_file.read_text(encoding="UTF-8"))

    return content


def save_json_file(json_file: Path, data: Union[List[Any], Dict[Any, Any]]) -> bool:
    """
    Saves the content of a list or dict to a json file.

    Args:
        json_file: The file to write the data into
        data:      The data to save

        Returns:
        True if data was valid
    """
    status = True
    if isinstance(data, (list, dict)):
        with json_file.open("w", encoding="UTF-8") as out_file:
            json.dump(data, out_file, indent=4)

    else:
        status = False

    return status


def save_yaml_file(yaml_file: Path, data: Union[List[Any], Dict[str, Any]]) -> bool:
    """
    Saves the content of a list or dict to a yaml file.

    Args:
        yaml_file: The file to write the data into
        data:      The data to save

    Returns:
        True if data was valid
    """
    status = True
    if isinstance(data, (list, dict)):
        with open(yaml_file, "w", encoding="UTF-8") as out_file:
            yaml.dump(data, out_file)

    else:
        status = False

    return status



"""
This is the fotoobo exception module.
"""

from typing import Any

from requests.exceptions import HTTPError


class GeneralException(Exception):
    """
    This is the generic fotoobo exception which can be used by any sub exception.
    """

    def __init__(self, message: str) -> None:
        """init"""
        self.message = message
        super().__init__(self.message)


class APIError(GeneralException):
    """Exception for errors with the network API"""

    def __init__(self, err: Any) -> None:
        """init"""
        self.http_status_codes = {
            200: "OK",
            400: "Bad Request",
            401: "Not Authorized",
            403: "Forbidden",
            404: "Resource Not Found",
            405: "Method Not Allowed",
            413: "Request Entity Too Large",
            424: "Failed Dependency",
            429: "Access temporarily blocked",
            500: "Internal Server Error",
        }
        # self.message = "general API Error"
        self.code = 999
        self.message = "unknown"

        if isinstance(err, HTTPError):
            self.code = err.response.status_code
            message = self.http_status_codes.get(self.code, "general API Error")
            self.message = f"HTTP/{str(self.code)} {message}"

        super().__init__(self.message)


class GeneralError(GeneralException):
    """
    The exception to raise if a general error occurred.
    The class does not have any methods as the only one (__init__) is inherited from its parent.
    Raise a GeneralError when it does not make sense to do further processing and the program should
    stop and exit.
    """


class GeneralWarning(GeneralException):
    """
    The exception to raise if a general warning occurred.
    The class does not have any methods as the only one (__init__) is inherited from its parent.
    Raise a GeneralWarning if a part of the program fails but it is safe to do further processing.
    """



"""
The FortiManager get commands
"""

import logging
from pathlib import Path
from typing import Dict, List, Union

import typer

from fotoobo.helpers import cli_path
from fotoobo.helpers.output import write_policy_to_html
from fotoobo.helpers.result import Result
from fotoobo.tools import fmg

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")

HELP_TEXT_HOST = "The FortiManager to access (must be defined in the inventory)."


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fmg get subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def adoms(
    host: str = typer.Argument(
        "fmg",
        help=HELP_TEXT_HOST,
        metavar="[host]",
    )
) -> None:
    """
    Get the FortiManager ADOM list.
    """
    result = fmg.get.adoms(host)
    result.print_result_as_table(title="FortiManager ADOMs", headers=["Name", "Version"])


@app.command()
def devices(
    host: str = typer.Argument(
        "fmg",
        help=HELP_TEXT_HOST,
        metavar="[host]",
    ),
    raw: bool = typer.Option(False, "-r", "--raw", help="Output raw data."),
) -> None:
    """
    Get the FortiManager devices list.

    In case of a cluster the 'Device Name' is the name of the cluster and the 'HA Nodes' holds the
    hostnames of the actual cluster nodes.
    """
    result = fmg.get.devices(host)
    if raw:
        result.print_raw()

    else:
        # Make a string from the list of ha nodes
        new_result = Result[Dict[str, Union[str, List[str]]]]()
        for key, value in result.all_results().items():
            value["ha_nodes"] = ", ".join(value["ha_nodes"])
            new_result.push_result(key, value)

        new_result.print_result_as_table(
            title="Fortinet devices",
            headers=["Device Name", "Version", "HA", "Platform", "Description", "HA nodes"],
        )


@app.command(no_args_is_help=True)
def policy(
    adom: str = typer.Argument(
        ...,
        help="The FortiManager ADOM to get the firewall policy from.",
        metavar="[adom]",
        show_default=False,
    ),
    policy_name: str = typer.Argument(
        ..., help="The name of the policy to get.", metavar="[policy]", show_default=False
    ),
    filename: Path = typer.Argument(
        ..., help="The filename to write the policy to.", metavar="[file]", show_default=False
    ),
    host: str = typer.Argument(
        "fmg",
        help=HELP_TEXT_HOST,
        metavar="[host]",
    ),
) -> None:
    """
    Get a FortiManager policy.
    """
    result = fmg.get.policy(host, adom, policy_name)
    write_policy_to_html(result.get_result(host), filename)


@app.command()
def version(host: str = typer.Argument("fmg", help=HELP_TEXT_HOST, metavar="[host]")) -> None:
    """
    Get the FortiManager version.
    """
    result = fmg.get.version(host)
    result.print_result_as_table(title="FortiManager Version", headers=["FortiManager", "Version"])



"""
FortiGate configuration check utility
"""

import logging
from pathlib import Path
from typing import Any, List

import typer

from fotoobo.exceptions import GeneralError, GeneralWarning
from fotoobo.fortinet.fortigate_config import FortiGateConfig
from fotoobo.fortinet.fortigate_config_check import FortiGateConfigCheck
from fotoobo.fortinet.fortigate_info import FortiGateInfo
from fotoobo.helpers.files import load_yaml_file
from fotoobo.helpers.result import Result

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


def check(config: Path, bundles: Path) -> Result[List[str]]:
    """
    The FortiGate configuration check

    Args:
        config:  The configuration to check (either a file or directory)
                 in case it's a directory all .conf files in it will be checked.
        bundles: The check bundle to check the configuration against

    Raises:
        GeneralWarning: GeneralWarning
        GeneralError:   GeneralError
    """
    files: List[Path] = []
    config = Path(config)

    if config.is_file():
        files.append(config)

    elif config.is_dir():
        log.debug("Given config is a directory")
        files = [file for file in config.iterdir() if file.suffix == ".conf"]

    else:
        log.error("No valid configuration file")

    if not files:
        log.warning("There are no configuration files to check")
        raise GeneralWarning("There are no configuration files to check")

    bundles = Path(bundles)
    if bundles.is_file():
        checks = load_yaml_file(bundles)

    else:
        log.error("No valid bundle file")
        raise GeneralError("No valid bundle file")

    total_results: int = 0
    result = Result[List[str]]()

    for file in files:
        try:
            fortigate_config = FortiGateConfig.parse_configuration_file(file)
            conf_check = FortiGateConfigCheck(fortigate_config, checks, result)

        except GeneralWarning as warn:
            log.warning(warn.message)
            continue

        conf_check.execute_checks()

        num_results = len(result.get_messages(fortigate_config.info.hostname))
        log.info("All checks in '%s' done with '%s' messages", file.name, num_results)
        total_results += num_results

    log.info("All checks done with '%s' messages", total_results)

    if total_results == 0:
        result.push_message("fotoobo", "There were no errors in the configuration file(s)")

    return result


def get(config: Path, scope: str = "", path: str = "") -> Result[FortiGateInfo]:
    """
    The FortiGate get configuration utility.

    Args:
        config: The configuration to get the information from (either a file or directory)
                In case it's a directory all .conf files in it will be checked.

    Returns:
        Configuration as result object

    Raises:
        GeneralWarning: GeneralWarning
    """
    files: List[Path] = []
    if config.is_file():
        files.append(config)

    elif config.is_dir():
        log.debug("Given config is a directory")
        files = [file for file in config.iterdir() if file.is_file() and file.suffix == ".conf"]

    if not files:
        log.warning("There are no configuration files")
        raise GeneralWarning("There are no configuration files")

    result = Result[Any]()

    for file in files:
        conf = FortiGateConfig.parse_configuration_file(file)
        output = conf.get_configuration(scope, path)
        result.push_result(conf.info.hostname, output)

    return result


def info(config: Path) -> Result[FortiGateInfo]:
    """
    The FortiGate configuration information utility.

    Args:
        config: The configuration to get the information from (either a file or directory)
                In case it's a directory all .conf files in it will be checked.

    Returns:
        FortiGate information as result object

    Raises:
        GeneralWarning: GeneralWarning
    """
    files: List[Path] = []
    if config.is_file():
        files.append(config)

    elif config.is_dir():
        log.debug("Given config is a directory")
        files = [file for file in config.iterdir() if file.is_file() and file.suffix == ".conf"]

    if not files:
        log.warning("There are no configuration files")
        raise GeneralWarning("There are no configuration files")

    result = Result[FortiGateInfo]()

    for file in files:
        conf = FortiGateConfig.parse_configuration_file(file)
        result.push_result(conf.info.hostname, conf.info)

    return result



"""
Devices class for storing device information
"""

import logging
import re
from pathlib import Path
from typing import Any, Dict, Optional

from fotoobo.exceptions import GeneralWarning
from fotoobo.fortinet.fortianalyzer import FortiAnalyzer
from fotoobo.fortinet.forticlientems import FortiClientEMS
from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.fortinet.fortimanager import FortiManager
from fotoobo.helpers.config import config
from fotoobo.helpers.files import load_yaml_file
from fotoobo.helpers.vault import Client

from .generic import GenericDevice

log = logging.getLogger("fotoobo")


class Inventory:
    """
    Represents an inventory full of Fortinet and generic devices. Instantiate the inventory by
    providing an inventory file in yaml format as described in the documentation at
    https://fotoobo.readthedocs.io/en/latest/usage/inventory.html
    """

    def __init__(self, inventory_file: Path) -> None:
        """Initialize the inventory

        Args:
            inventory file: The Path object to the inventory file.
        """
        log.debug("Initialize the fotoobo inventory")
        self._inventory_file: Path = inventory_file
        self._globals: Dict[str, Any] = {
            "fortianalyzer": {},
            "forticlientems": {},
            "fortigate": {},
            "fortimanager": {},
        }
        self.assets: Dict[str, Any] = {}
        self.vault_data: Dict[str, Dict[str, str]] = {}

        # Load assets from inventory file
        self._load_inventory()

        # Load credentials from a configured Hashicorp vault and enrich assets
        if config.vault:
            self._load_data_from_vault(config.vault)
            self._replace_with_vault_data()

        # Create object for FortiGates
        self.fortigates = {
            name: asset for (name, asset) in self.assets.items() if isinstance(asset, FortiGate)
        }

    def get(
        self,
        name: Optional[str] = None,
        type: Optional[str] = None,  # pylint: disable=redefined-builtin
    ) -> Dict[str, Any]:
        """
        Get a dictionary of assets from the inventory with given 'filters'.

        Args:
            name: Which asset to get. If blank get every asset. Wildcard * is supported in any
                  position. E.g. "FortiGate", "*Gate", "Forti*ate", "Forti*".
            type: Which asset type to get. If blank get every asset.

        Returns:
            Dictionary of assets with name as key

        Raises:
            GeneralWarning: If no asset was found in the inventory that matches name or type
        """
        log.debug("Getting assets with name '%s' and type '%s'", name, type)
        name_pattern = f"^{name}$".replace("*", ".*")
        assets = {}

        for _name, _asset in self.assets.items():
            if not name and not type:
                assets[_name] = _asset

            elif not name and getattr(_asset, "type", None) == type:
                assets[_name] = _asset

            elif name and re.match(name_pattern, _name):
                if (type and getattr(_asset, "type", None) == type) or not type:
                    assets[_name] = _asset

        if not assets:
            raise GeneralWarning(
                f"no asset of type '{type}' and name '{name}' was found in the inventory"
            )

        return assets

    def get_item(
        self,
        name: str = "",
        type: Optional[str] = None,  # pylint: disable=redefined-builtin
    ) -> Any:
        """
        Get a single asset from the inventory. To be sure the item has the correct type you can
        specify the desired type and a GeneralWarning is raised if the type does not match.

        Args:
            name: Which exact asset to get.
            type: If you specify a type it is checked if it matches the type of the asset

        Returns:
            Inventory asset

        Raises:
            GeneralWarning: If the asset was not found in the inventory
            GeneralWarning: If the specified type does not match the type of the item
        """
        log.debug("Getting asset with name '%s'", name)

        try:
            asset = self.assets[name]

        except KeyError as error:
            raise GeneralWarning(
                f"Asset with name '{name}' was not found in the inventory"
            ) from error

        if type and type != asset.type:
            raise GeneralWarning(f"Asset with name '{name}' is not of type '{type}'")

        return asset

    def _load_inventory(self) -> None:
        """
        Load the inventory from a file given
        """
        log.debug("Load the assets from the inventory file")
        # Expand user home shortcuts in the inventory file path
        expanded_inventory_file = self._inventory_file.expanduser()
        log.debug("Loading assets from '%s'", expanded_inventory_file)
        inventory_raw: Dict[str, Any] = dict(load_yaml_file(expanded_inventory_file) or {})

        # Set the globals
        # This has to be done before the looping over all inventory items so that the globals are
        # already set during looping
        if "globals" in inventory_raw:
            self._set_globals(inventory_raw["globals"])

        for name, asset in inventory_raw.items():
            # skip globals
            if name == "globals":
                continue

            # enrich the raw asset data with the globals
            if asset.get("type", None) in self._globals:
                asset = {**self._globals[asset["type"]], **asset}

            if asset.get("type", "") == "fortigate":
                try:
                    self.assets[name] = FortiGate(**asset)

                except GeneralWarning as err:
                    log.warning("%s: %s", name, err)

            elif asset.get("type", "") == "forticlientems":
                self.assets[name] = FortiClientEMS(**asset)

            elif asset.get("type", "") == "fortianalyzer":
                self.assets[name] = FortiAnalyzer(**asset)

            elif asset.get("type", "") == "fortimanager":
                self.assets[name] = FortiManager(**asset)

            else:
                self.assets[name] = GenericDevice(**asset)

    # pylint: disable=pointless-string-statement
    """
        # create asset object using the match and case keywords
        match asset_type:
            case "fortigate":
                self.assets[name] = FortiGate(**asset)
            case "forticlientems":
                self.assets[name] = FortiClientEMS(**asset)
            case "fortianalyzer":
                self.assets[name] = FortiAnalyzer(**asset)
            case "fortimanager":
                self.assets[name] = FortiManager(**asset)
            case _:
                self.assets[name] = GenericDevice(**asset)
    """

    def _load_data_from_vault(self, vault_dict: Dict[str, Any]) -> None:
        """Load the credentials from a vault

        Args:
            vault_dict: The Dictionary with the vault configuration
        """
        log.debug("Loading credentials from vault '%s'", vault_dict["url"])
        vault = Client(**vault_dict)
        data = vault.get_data()
        self.vault_data = data.get("ok", {}).get("data", {}).get("data", {})

    def _replace_with_vault_data(self) -> None:
        """Replace asset attributes with data from the vault"""
        for name, asset in self.assets.items():
            for attribute in dir(asset):
                if not attribute.startswith("_"):
                    if getattr(asset, attribute) == "VAULT":
                        try:
                            setattr(asset, attribute, self.vault_data[name][attribute])
                            log.debug(
                                "Vault attribute '%s.%s' replaced successfully", name, attribute
                            )

                        except KeyError:
                            log.warning("Vault attribute '%s.%s' not found", name, attribute)

    def _set_globals(self, data: Dict[str, Any]) -> None:
        """
        Set some defaults for device types

        Args:
            data: defaults by device where key is the device type and value defines its defaults
        """
        for key, value in data.items():
            self._globals[key] = value



"""
The helper module

Here you'll find some useful functions and classes for interacting with fotoobo. Mainly there are
configuration, logging and input/output functions.
Put everything into this module what does not belong directly to any other module.
"""

from typing import List

cli_path: List[str] = []



"""
This is the main project entry point.

When invoking fotoobo it starts the main() function in this file. Its main purpose is to start the
typer cli with app().
The second task is to catch all exceptions and print a friendly message on the screen instead of
a traceback. The traceback is written to a traceback.log file in the local directory for debug
purposes.

The exceptions in the main functions cannot be tested because the test-method for typer always
exits with its own exit codes (1 or 2) when an exception rises.
"""

import sys
import traceback

from fotoobo.cli.main import app
from fotoobo.exceptions import APIError, GeneralError, GeneralWarning


def main() -> None:
    """
    This is the main function
    """
    try:
        app()

    except GeneralWarning as warn:  # pragma: no cover
        print(f"Warning: {warn.message}")
        sys.exit(30)

    except (GeneralError, APIError) as err:  # pragma: no cover
        print(f"Error: {err.message}")
        sys.exit(40)

    except Exception:  # pylint: disable=broad-except # pragma: no cover
        print("oops, something did not work as expected. See traceback.log for more info")
        with open("traceback.log", "w", encoding="UTF-8") as exc_file:
            traceback.print_exc(file=exc_file)
        sys.exit(50)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
FortiAnalyzer Class
"""

import logging
from typing import Any

from .fortimanager import FortiManager

log = logging.getLogger("fotoobo")


class FortiAnalyzer(FortiManager):
    """
    Represents one FortiAnalyzer (digital twin)

    This class is inherited from FortiManager. Only the type of the object differs.
    """

    def __init__(self, hostname: str, username: str, password: str, **kwargs: Any) -> None:
        """
        Set some initial parameters.

        Args:
            hostname: The hostname of the FortiGate to connect to
            username: Username
            password: Password
            **kwargs: See Fortinet class for available arguments
        """
        super().__init__(hostname, username, password, **kwargs)
        self.type = "fortianalyzer"

    def get_version(self) -> str:
        """
        Get FortiAnalyzer version

        Returns:
            FortiAnalyzer version
        """
        faz_version = super().get_version()
        return faz_version



"""
FortiGate Class
"""

import logging
from typing import Any, Dict, Optional

import requests

from fotoobo.exceptions import APIError, GeneralWarning

from .fortinet import Fortinet

log = logging.getLogger("fotoobo")


class FortiGate(Fortinet):
    """
    Represents one FortiGate (digital twin)
    """

    def __init__(
        self,
        hostname: str = "",
        token: str = "",
        **kwargs: Dict[str, str],
    ) -> None:
        """
        Set some initial parameters.

        Args:
            hostname: The hostname of the FortiGate to connect to
            token:    API access token from the FortiGate
            **kwargs: See Fortinet class for available arguments
        """
        if not hostname:
            raise GeneralWarning("No hostname specified")

        super().__init__(hostname=hostname, **kwargs)
        self.api_url = f"https://{self.hostname}:{self.https_port}/api/v2"
        self.token = token
        self.type = "fortigate"

    def api(  # pylint: disable=too-many-arguments
        self,
        method: str,
        url: str = "",
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        payload: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
    ) -> requests.models.Response:
        """
        API request to a FortiGate device.
        It uses the super.api method but it has to enrich the payload in post requests with the
        needed session key.

        Args:
            method:  Request method from [get, post]
            url:     Rest API URL to request data from
            params:  Dictionary with parameters (if needed)
            payload: JSON body for post requests (if needed)
            timeout: The requests read timeout

        Returns:
            Response from the request
        """
        self.session.headers.update({"Authorization": f"Bearer {self.token}"})
        return super().api(
            method, url, payload=payload, params=params, timeout=timeout, headers=headers
        )

    def backup(self, timeout: int = 10) -> str:
        """
        Get the configuration backup from a FortiGate.

        Args:
            timeout: Timeout in sec to wait for the response

        Returns:
            Configuration backup as text
        """
        data = self.api(
            "get", "monitor/system/config/backup", params={"scope": "global"}, timeout=timeout
        )
        return data.text

    def get_version(self) -> str:
        """
        Get FortiGate version

        Returns:
            FortiGate version
        """
        fgt_version: str = ""

        try:
            response = self.api("get", "monitor/system/status")

        except APIError as err:
            log.warning("'%s' returned: '%s'", self.hostname, err.message)
            raise GeneralWarning(f"{self.hostname} returned: {err.message}") from err

        fgt_version = response.json().get("version", "unknown")
        return fgt_version



"""
The logging helper class is used if you want to write logs to the screen or a file. It is set up
to do pretty printing screen output with rich and writing "normal" text to a logfile.
Basic logging options can be configured in the main configuration file. See config_helper.py
"""

import logging
import logging.config
import os
import pwd
import socket
from datetime import datetime
from logging.handlers import RotatingFileHandler, SysLogHandler
from syslog import (
    LOG_AUTH,
    LOG_CRIT,
    LOG_DEBUG,
    LOG_ERR,
    LOG_INFO,
    LOG_USER,
    LOG_WARNING,
)
from typing import Optional

from rich.logging import RichHandler

from fotoobo.exceptions import GeneralError, GeneralWarning
from fotoobo.helpers.config import config
from fotoobo.helpers.files import load_yaml_file

logger = logging.getLogger("fotoobo")
audit_logger = logging.getLogger("audit")


class SysLogFormatter(logging.Formatter):
    """
    Handles syslog formats
    """

    def __init__(self, facility: int, fmt: str = "%(message)s", datefmt: Optional[str] = None):
        """
        Initialize the logger

        Args:
            facility:
        """
        self.facility = facility

        # Need to do this like this, to have the GitLab pipeline work
        # See https://stackoverflow.com/questions/4399617/python-os-getlogin-problem
        self.user = pwd.getpwuid(os.getuid())[0]
        self.hostname = socket.gethostname()
        super().__init__(fmt=fmt, datefmt=datefmt)

    def format(self, record: logging.LogRecord) -> str:
        """
        Format the message

        Args:
            record:

        Returns:
            A string
        """
        if record.levelname == "DEBUG":
            level: int = LOG_DEBUG

        elif record.levelname == "INFO":
            level = LOG_INFO

        elif record.levelname == "WARNING":
            level = LOG_WARNING

        elif record.levelname == "ERROR":
            level = LOG_ERR

        elif record.levelname == "CRITICAL":
            level = LOG_CRIT

        else:
            raise NotImplementedError(f"Loglevel {record.levelname} cannot be processed!")

        msg_id = "AUDIT" if self.facility == LOG_AUTH else "-"

        prival = self.facility + level
        timestamp = datetime.fromtimestamp(record.created).astimezone().isoformat()
        return (
            " ".join(
                [
                    # Syslog header parts
                    f"<{prival}>1",
                    timestamp,
                    f"{self.hostname}",
                    "fotoobo",
                    f"{record.process}",
                    msg_id,
                    "-",
                    # Syslog message parts
                    f"username={self.user}",
                ]
            )
            + " "
            + super().format(record)
        )


class Log:
    """
    The logger class for log control.
    """

    @staticmethod
    def configure_logging(quiet: Optional[bool], log_level: Optional[str]) -> None:
        """
        This function will configure the logging for fotoobo

        Args:
            quiet:     Disables the console logging
            log_level: The desired log_level (given by CLI argument)

        Raises:
            GeneralError: On unrecoverable errors (usually on non-existing/empty or
                          invalid logging configuration file
        """
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        # Make some defaults for (some chatty) external libraries
        logging.getLogger("requests").setLevel(logging.CRITICAL)
        logging.getLogger("urllib3").setLevel(logging.CRITICAL)
        logging.getLogger("urllib3.connectionpool").setLevel(logging.CRITICAL)

        # If we have a logging configuration file, use this to configure logging
        if config.logging and config.logging.get("log_configuration_file"):
            if logging_config := load_yaml_file(config.logging["log_configuration_file"]):
                try:
                    logging.config.dictConfig(dict(logging_config))

                except ValueError as error:
                    raise GeneralError(f"Cannot configure logging: {str(error)}") from error

                except TypeError as error:
                    raise GeneralError(f"Cannot configure logging: {str(error)}") from error

                except AttributeError as error:
                    raise GeneralError(f"Cannot configure logging: {str(error)}") from error

                except ImportError as error:
                    raise GeneralError(f"Cannot configure logging: {str(error)}") from error

            # If the configuration file cannot be found or is empty
            else:
                raise GeneralError(
                    "Cannot configure logging: Configuration file "
                    f'"{config.logging["log_configuration_file"]}" not found or empty!'
                )

        # If there is no logging configuration file, use the basic logging configuration
        else:
            # The log level given as command line option has precedence over anything other
            if log_level and log_level not in [
                "CRITICAL",
                "ERROR",
                "WARNING",
                "INFO",
                "DEBUG",
            ]:
                raise GeneralWarning(f"Loglevel {log_level} not known")

            # If nothing is configured in the config file use a simple basic config
            if not config.logging:
                # Configure logger "fotoobo"
                logger.setLevel(log_level or logging.WARNING)
                console_handler = RichHandler(markup=True)
                console_handler.setFormatter(
                    logging.Formatter(fmt="[grey37]%(name)s[/] %(message)s", datefmt="[%X]")
                )
                logger.addHandler(console_handler)

                # Configure audit_logger
                audit_logger.disabled = True

            # If there is a configuration in the config file
            else:
                # Only configure something if there is a logging configuration and logging is
                # requested
                if config.logging:
                    log_level = log_level if log_level else config.logging["level"]
                    logger.setLevel(log_level)
                    logging.getLogger("requests").setLevel(log_level)
                    logging.getLogger("urllib3").setLevel(log_level)
                    logger.handlers = []

                    # Configure console logging
                    if "log_console" in config.logging:
                        console_handler = RichHandler(markup=True)
                        console_handler.setFormatter(
                            logging.Formatter(fmt="[grey37]%(name)s[/] %(message)s", datefmt="[%X]")
                        )
                        logger.addHandler(console_handler)

                    if "log_file" in config.logging:
                        file_handler = RotatingFileHandler(
                            filename=config.logging["log_file"]["name"],
                            maxBytes=10485760,  # 10 MByte
                            backupCount=3,
                        )
                        file_handler.setFormatter(
                            logging.Formatter(
                                "%(asctime)s - %(name)s - %(levelname)s - %(filename)s:%(lineno)d: "
                                "%(message)s"
                            )
                        )
                        logger.addHandler(file_handler)

                    if "log_syslog" in config.logging:
                        try:
                            syslog_handler = SysLogHandler(
                                address=(
                                    config.logging["log_syslog"]["host"],
                                    int(config.logging["log_syslog"]["port"]),
                                ),
                                socktype=(
                                    socket.SOCK_STREAM
                                    if config.logging["log_syslog"]["protocol"] == "TCP"
                                    else socket.SOCK_DGRAM
                                ),
                            )

                        except (OSError, socket.gaierror) as error:
                            raise GeneralError(
                                f"Cannot configure SysLog logging: {str(error)}"
                            ) from error

                        syslog_handler.setFormatter(SysLogFormatter(LOG_USER))
                        logger.addHandler(syslog_handler)

                # Configure audit logging
                if config.audit_logging:
                    audit_logger.setLevel(logging.INFO)
                    if "log_file" in config.audit_logging:
                        audit_file_handler = RotatingFileHandler(
                            filename=config.audit_logging["log_file"]["name"],
                            maxBytes=10485760,  # 10 MByte
                            backupCount=3,
                        )
                        audit_file_handler.setFormatter(
                            logging.Formatter(
                                fmt="%(asctime)s - AUDIT - %(filename)s:%(lineno)d: %(message)s"
                            )
                        )
                        audit_logger.addHandler(audit_file_handler)

                    if "log_syslog" in config.audit_logging:
                        try:
                            audit_syslog_handler = SysLogHandler(
                                address=(
                                    config.audit_logging["log_syslog"]["host"],
                                    int(config.audit_logging["log_syslog"]["port"]),
                                ),
                                facility=LOG_AUTH,
                                socktype=(
                                    socket.SOCK_STREAM
                                    if config.audit_logging["log_syslog"]["protocol"] == "TCP"
                                    else socket.SOCK_DGRAM
                                ),
                            )

                        except (OSError, socket.gaierror) as error:
                            raise GeneralError(
                                f"Cannot configure SysLog logging: {str(error)}"
                            ) from error

                        audit_syslog_handler.setFormatter(SysLogFormatter(LOG_AUTH))
                        audit_logger.addHandler(audit_syslog_handler)

                else:
                    audit_logger.disabled = True

            if quiet:
                logger.disabled = True

    @staticmethod
    def audit(message: str) -> None:
        """
        Create an audit log message
        """
        audit_logger.info(message)


logger.audit = Log.audit  # type: ignore



"""
FortiClient EMS get module
"""

import logging
from typing import Dict, List

from fotoobo.fortinet.forticlientems import FortiClientEMS
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def version(host: str) -> Result[str]:
    """
    ems get version

    Args:
        host: Host defined in inventory

    Returns:
        Version data in a dict with keys: host, version
    """
    result = Result[str]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    log.debug("FortiClient EMS get version ...")
    ems.login()
    ems_version = ems.get_version()
    result.push_result(host, f"v{ems_version}")
    return result


def workgroups(host: str, custom: bool = False) -> Result[List[Dict[str, str]]]:
    """
    ems get workgroups

    Args:
        host:   Host defined in inventory
        custom: If true it only returns custom groups

    Returns:
        Workgroups data in a list of dict with keys: name, id, total_devices
    """
    result = Result[List[Dict[str, str]]]()
    inventory = Inventory(config.inventory_file)
    ems: FortiClientEMS = inventory.get_item(host, "forticlientems")
    log.debug("FortiClient EMS get workgroups ...")
    ems.login()
    raw_data = ems.api("get", f"/workgroups/index?custom={custom}").json()["data"]

    for entry in raw_data:
        result.push_result(entry["name"], entry["total_devices"])

    return result



"""
fotoobo - make IT easy

This is fotoobo, the mighty Fortinet toolbox for managing your Fortinet environment. It is meant
to be extendable to your needs. It's most likely the swiss army knife for Fortinet infrastructure.
"""

from .fortinet import FortiAnalyzer, FortiClientEMS, FortiGate, FortiManager

__all__ = ["FortiAnalyzer", "FortiClientEMS", "FortiGate", "FortiManager"]
__version__: str = "2.2.0"



"""
The FortiClient EMS commands
"""

import logging

import typer

from fotoobo.cli.ems import get_commands as get
from fotoobo.cli.ems import monitor_commands as monitor
from fotoobo.helpers import cli_path

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The ems subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


app.add_typer(get.app, name="get", help="FortiClient EMS get commands.")
app.add_typer(monitor.app, name="monitor", help="FortiClient EMS monitor commands.")



"""
The fotoobo greeting utility

This function is a hidden fotoobo command which is meant for testing purposes. It has no functional
effect. Use and modify it whenever you want. But be sure to also write the tests for it.
"""

import logging

import typer

log = logging.getLogger("fotoobo")


def greet(name: str, bye: bool, log_enabled: bool) -> None:
    """
    This is the hidden Greeting function.
    It allows you to greet someone with different colors in different languages.

    Args:
        name:
        bye:
        log_enabled:
    """
    log.debug("Local logging option is: '%s'", log_enabled)
    greeting = f"{typer.style('🌼Aloha🌼', fg=typer.colors.GREEN)}"
    greeting += f", {typer.style('⚽Hola⚽', fg=typer.colors.YELLOW)}"
    greeting += f", {typer.style('✨Bonjour✨', fg=typer.colors.BLUE)}"
    greeting += f", {typer.style('⚡Hallo⚡', fg=typer.colors.RED)}"
    greeting += f", {typer.style('☀Ciao☀', fg=typer.colors.BRIGHT_GREEN)}"
    greeting += f", {typer.style('🌟Konnichiwa🌟', fg=typer.colors.BRIGHT_YELLOW)}"
    greeting += f", {typer.style('🎉Howdy-doody🎉', fg=typer.colors.BRIGHT_BLUE)}!"

    if name:
        greeting = f"Hi {typer.style(name, bold=True)}, {greeting}"

    typer.echo(greeting)

    if bye:
        typer.echo(typer.style("Good Bye", fg=typer.colors.BLACK, bg=typer.colors.WHITE))



"""
FortiGate check hamaster utility
"""

import concurrent.futures
import logging
from typing import Dict, Tuple

from rich.progress import Progress

from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def hamaster(host: str) -> Result[str]:  # pylint: disable=too-many-locals
    """FortiGate check hamaster.

    This method first gets all the devices from a FortiManager to find all the managed FortiGates
    clusters. Then it queries every cluster for its HA status directly to its shared management ip.
    Be aware that the device names in FortiManager must match the names in the inventory because we
    search for the devices we found in Fortimanager in our inventory to connect to to them.

    Args:
        host: The FortiManager host from the inventory to get the device list from. If you omit
              host, it will run over the default FortiManager (fmg).

    Returns:
        The Result object with all the results
    """

    def _get_single_status(name: str, fgt: FortiGate) -> Tuple[str, str]:
        """Get the HA master status from a FortiGate.

        This private method is used for multithreading. It only queries one single FortiGate for its
        HA master status and returns it.

        Args:
            name: The name of the FortiGate (as defined in the inventory)
            fgt:  The FortiGate object to query

        Returns:
            name:   The name of the FortiGate (as defined in the inventory)
            status: The HA status of the FortiGate (fgt)
        """
        response = fgt.api("get", "/monitor/system/ha-checksums")
        ha_checksums = response.json()
        status: str = "is not the expected master"

        for node in ha_checksums["results"]:
            if node["serial_no"] == ha_checksums["serial"]:
                if node["is_root_master"] == 1:
                    status = "ok"

        return name, status

    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    payload = {
        "method": "get",
        "params": [
            {
                "loadsub": 1,
                "sortings": [{"build": 1}],
                "option": "object member",
                "url": "/dvmdb/device",
            }
        ],
    }
    fmg.login()
    response = fmg.api("post", payload=payload)
    fmg.logout()
    result = Result[str]()
    fgts: Dict[str, FortiGate] = {}

    for device in response.json()["result"][0]["data"]:
        if device["ha_mode"] == 1:
            expected_master = ""
            highest = 0
            # Loop over all the cluster nodes and find the node with the hightest priority
            for node in device["ha_slave"]:
                if node["prio"] > highest:
                    highest = node["prio"]
                    expected_master = node["name"].lower()

            try:
                fortigate: FortiGate = inventory.assets[expected_master]
                # Replace the FortiGates Hostname with the cluster IP address. In case of a HA
                # failover the designated master may not be reachable so we connect to the cluster
                # IP address.
                fortigate.hostname = device["ip"]
                fgts[expected_master] = fortigate

            # There is a KeyError if a designated cluster master is not defined in the inventory
            except KeyError:
                log.debug("Device '%s' not found in inventory", expected_master)
                result.push_result(expected_master, "not found in inventory")

    with Progress() as progress:
        task = progress.add_task("Getting FortiGate HA status...", total=len(fgts))
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            futures = []
            for name, fgt in fgts.items():
                futures.append(executor.submit(_get_single_status, name, fgt))

            for future in concurrent.futures.as_completed(futures):
                name, status = future.result()
                result.push_result(name, status)
                progress.update(task, advance=1)

    return result



"""
The fotoobo converter utility
"""

import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from fotoobo.fortinet.convert import CheckpointConverter
from fotoobo.helpers.files import create_dir
from fotoobo.helpers.result import Result

log = logging.getLogger("fotoobo")


def checkpoint(
    checkpoint_assets: Union[List[Any], Dict[str, Any], None],
    obj_type: str,
    conversion_id: str,
    cache_dir: Optional[Path] = None,
) -> Result[List[Any]]:
    """
    This function lets you convert Checkpoint assets into Fortinet syntax.

    Args:
        checkpoint_assets: The assets to convert
        obj_type:          The type of asset to convert
        conversion_id:     An ID to the conversion. Needed to cache different conversion tasks
        cache_dir:         The cache directory to use. Defaults to None.
    """
    log.info("Start converting checkpoint assets of type '%s'", obj_type)
    result = Result[List[Any]]()

    if cache_dir:
        create_dir(cache_dir)
        cache_file = Path(cache_dir, conversion_id)

    else:
        cache_file = None

    converter = CheckpointConverter(checkpoint_assets, cache_file)
    result.push_result("fortinet_assets", converter.convert(obj_type))
    return result



"""The powerful convert module

Please refer to docs_legacy/convert_checkpoint_mappings.drawio.svg to see how a Checkpoint asset is
going to be converted into a Fortinet asset.

The Fortinet assets are referenced by their names. Even if we set an uuid we cannot use the uuid to
access the asset. The uuid is only supported for network assets and not for service assets.
"""

import copy
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional

from fotoobo.exceptions import GeneralError
from fotoobo.helpers.files import load_json_file, save_json_file

log = logging.getLogger("fotoobo")


class CheckpointConverter:
    """
    The Checkpoint converter class
    """

    def __init__(self, assets: Any, cache_file: Optional[Path] = None) -> None:
        self.assets = copy.deepcopy(assets)
        self.converted: List[Any] = []
        self.supported_types = [
            "hosts",
            "networks",
            "address_ranges",
            "groups",
            "services_icmp",
            "services_icmp6",
            "services_tcp",
            "services_udp",
            "service_groups",
        ]
        self.cache_file = cache_file

    def convert(self, obj_type: str, bulk_size: int = 100) -> List[Any]:
        """
        Convert Checkpoint configuration objects into Fortinet syntax.

        Args:
            obj_type:   Define the type of objects to convert
            bulk_size:  The bulk size to generate

        Returns:
            The converted assets
        """
        # fix for bulk problem with groups:
        # It seems that adding groups in bulk mode does not work correctly. If one of the entries
        # in a bulk set is invalid all of them are not added with the same error.
        if obj_type == "groups":
            bulk_size = 1

        if obj_type not in self.supported_types:
            raise GeneralError(f"type '{obj_type}' is not supported to convert")

        if obj_type not in self.assets:
            raise GeneralError(f"type '{obj_type}' is not present in the infile")

        log.debug("Converting asset type '%s'", obj_type)
        log.debug("Found '%s' asset(s)", len(self.assets[obj_type]))
        assets = getattr(self, "_convert_" + obj_type)()

        # Now do the caching if cache_file is given
        if self.cache_file:
            cache: List[Any] = []
            if self.cache_file.is_file():
                log.debug("Found cache file '%s'", self.cache_file)
                cache = list(load_json_file(self.cache_file) or [])

            save_json_file(self.cache_file, assets)
            assets = [asset for asset in assets if asset not in cache]
            log.debug("Found '%s' new asset(s) not in cache", len(assets))

        # Now do the bulking
        i = 0
        params = []
        while assets:
            params.append(assets.pop(0))
            i += 1
            if i == bulk_size or not assets:
                converted = {
                    "method": "set",
                    "params": params,
                    "session": "",
                    "id": 1,
                }
                self.converted.append(converted)
                i = 0
                params = []

        return self.converted

    def _convert_hosts(self) -> Any:
        """
        Convert Checkpoint host objects

        Returns:
            Converted Fortinet objects
        """
        objects = self.assets["hosts"]
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            param = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "subnet": f"{obj['ipv4-address']}/32",
                    "uuid": obj["uid"],
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/address/{url_name}",
            }
            assets.append(param)

        return assets

    def _convert_networks(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint network objects

        Returns:
            Converted Fortinet objects
        """
        objects = self.assets["networks"]
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            param = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "subnet": f"{obj['subnet4']}/{obj['mask-length4']}",
                    "uuid": obj["uid"],
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/address/{url_name}",
            }
            assets.append(param)

        return assets

    def _convert_address_ranges(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint address range objects

        Returns:
            Converted Fortinet objects
        """
        objects = self.assets["address_ranges"]
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            param = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "type": "iprange",
                    "start-ip": f"{obj['ipv4-address-first']}",
                    "end-ip": f"{obj['ipv4-address-last']}",
                    "uuid": obj["uid"],
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/address/{url_name}",
            }
            assets.append(param)

        return assets

    def _convert_groups(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint group objects

        Returns:
            Converted Fortinet objects
        """
        objects = copy.deepcopy(self.assets["groups"])
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            forti_members = []
            for uid in obj["members"]:
                found = False
                for obj_type in ["hosts", "networks", "address_ranges", "groups"]:
                    for net_object in self.assets[obj_type]:
                        if net_object["uid"] == uid:
                            forti_members.append(net_object["name"])
                            found = True

                if not found:
                    log.error("Object with uid '%s' not found", uid)

            param = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "member": forti_members,
                    "uuid": obj["uid"],
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/addrgrp/{url_name}",
            }
            if forti_members:
                assets.append(param)

            else:
                log.error("Network group '%s' is empty", url_name)

        return assets

    def _convert_services_icmp(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint services_icmp objects

        Returns:
            Converted Fortinet objects
        """
        objects = self.assets["services_icmp"]
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            param: Dict[str, Any] = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "protocol": "ICMP",
                    "icmptype": obj["icmp-type"],
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/service/custom/{url_name}",
            }
            if "icmp-code" in obj:
                param["data"]["icmpcode"] = obj["icmp-code"]

            assets.append(param)

        return assets

    def _convert_services_icmp6(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint services_icmp6 objects

        Returns:
            Converted Fortinet objects
        """
        objects = self.assets["services_icmp6"]
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            param: Dict[str, Any] = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "protocol": "ICMP6",
                    "icmptype": obj["icmp-type"],
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/service/custom/{url_name}",
            }
            assets.append(param)

        return assets

    def _convert_services_tcp(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint services_tcp objects

        Possible values for "port" in the Checkpoint assets are:
        - "2000" is a single port which will be "2000"
        - "3000-3500" is a port range which will be "3000-4000"
        - ">4000" will be converted to "4001-65535"
        - "<5000" will be converted to "1-4999"

        Returns:
            Converted Fortinet objects
        """
        objects = self.assets["services_tcp"]
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            if obj["port"][0] == ">":
                port_range = str(int(obj["port"][1:]) + 1) + "-65535"

            elif obj["port"][0] == "<":
                port_range = "1-" + str(int(obj["port"][1:]) - 1)

            else:
                port_range = obj["port"]

            param: Dict[str, Any] = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "protocol": "TCP/UDP/SCTP",
                    "tcp-portrange": f"{port_range}",
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/service/custom/{url_name}",
            }
            if not obj["use-default-session-timeout"]:
                if (ttl := obj["session-timeout"]) < 300:  # FortiGate does not allow ttl < 300
                    ttl = 300

                param["data"]["session-ttl"] = str(ttl)

            assets.append(param)

        return assets

    def _convert_services_udp(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint services_udp objects
        Even though the UDP services are almost identical to the TCP services we use its own
        convert function. This adds some redundancy but removes complexity and segregates duties.

        For possible values for the Checkpoint "port" field and how they will get converted refer
        to the _convert_services_tcp function.

        Returns:
            Converted Fortinet objects
        """
        objects = self.assets["services_udp"]
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")
            if obj["port"][0] == ">":
                port_range = str(int(obj["port"][1:]) + 1) + "-65535"

            elif obj["port"][0] == "<":
                port_range = "1-" + str(int(obj["port"][1:]) - 1)

            else:
                port_range = obj["port"]

            param: Dict[str, Any] = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "protocol": "TCP/UDP/SCTP",
                    "udp-portrange": f"{port_range}",
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/service/custom/{url_name}",
            }
            if not obj["use-default-session-timeout"]:
                if (ttl := obj["session-timeout"]) < 300:  # FortiGate does not allow ttl < 300
                    ttl = 300

                param["data"]["session-ttl"] = str(ttl)

            assets.append(param)

        return assets

    def _convert_service_groups(self) -> List[Dict[str, Any]]:
        """
        Convert Checkpoint service group objects

        Returns:
            Converted Fortinet objects
        """
        objects = copy.deepcopy(self.assets["service_groups"])
        assets = []
        while objects:
            obj = objects.pop(0)
            url_name = obj["name"].replace("/", "\\/")

            # patch for special group "gIntegrity_Server"
            # This one cannot be converted as the members are not supported by FortiManager.
            # Remove the group "gIntegrity_Server" from the Checkpoint global policy (it seems not
            # to be in use). Then this patch can be removed.
            if url_name == "gIntegrity_Server":
                continue  # pragma: no cover

            forti_members = []
            for uid in obj["members"]:
                found = False
                for obj_type in [
                    "services_tcp",
                    "services_udp",
                    "services_icmp",
                    "services_icmp6",
                    "service_groups",
                ]:
                    for net_object in self.assets[obj_type]:
                        if net_object["uid"] == uid:
                            forti_members.append(net_object["name"])
                            found = True

                if not found:
                    log.error("Object with uid '%s' not found", uid)

            param = {
                "data": {
                    "comment": obj["comments"],
                    "name": obj["name"],
                    "member": forti_members,
                },
                "url": f"/pm/config/{{adom}}/obj/firewall/service/group/{url_name}",
            }
            if forti_members:
                assets.append(param)

            else:
                log.error("service group '%s' is empty", url_name)

        return assets



"""
The FortiGate get commands
"""

import logging
from pathlib import Path

import typer

from fotoobo.helpers.config import config
from fotoobo.inventory.inventory import Inventory
from fotoobo.tools import fgt

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.command(no_args_is_help=True)
def check(
    configuration: Path = typer.Argument(
        ...,
        help="The FortiGate configuration file or directory.",
        metavar="[config]",
        show_default=False,
    ),
    bundles: Path = typer.Argument(
        ...,
        help="Filename of the file containing the check bundles.",
        metavar="[bundles]",
        show_default=False,
    ),
    smtp_server: str = typer.Option(
        None,
        "--smtp",
        help="The smtp configuration from the inventory.",
        metavar="[server]",
        show_default=False,
    ),
) -> None:
    """
    Check one or more FortiGate configuration files.
    """
    inventory = Inventory(config.inventory_file)
    result = fgt.config.check(configuration, bundles)

    if smtp_server:
        if smtp_server in inventory.assets:
            result.send_messages_as_mail(
                inventory.assets[smtp_server],
                count=True,
                command=True,
            )

        else:
            log.warning("SMTP server '%s' not in found in inventory.", smtp_server)

    result.print_messages()


@app.command(no_args_is_help=True)
def get(
    configuration: Path = typer.Argument(
        ...,
        help="The FortiGate configuration file or directory.",
        metavar="[config]",
        show_default=False,
    ),
    scope: str = typer.Argument(
        ..., help="Scope of the configuration ('global' or 'vdom')", metavar="[scope]"
    ),
    path: str = typer.Argument("/", help="Configuration path", metavar="[path]"),
) -> None:
    """Get configuration or parts of it from one or more FortiGate configuration files"""
    result = fgt.config.get(configuration, scope, path)
    result.print_raw()


@app.command(no_args_is_help=True)
def info(
    configuration: Path = typer.Argument(
        ...,
        help="The FortiGate configuration file or directory.",
        metavar="[config]",
        show_default=False,
    ),
    as_list: bool = typer.Option(
        False, "--list", "-l", help="Print the result as a list instead of separate blocks."
    ),
) -> None:
    """
    Get the information from one or more FortiGate configuration files.
    """
    result = fgt.config.info(configuration)

    if as_list:
        info_dicts = []
        for _, _info in result.all_results().items():
            info_dicts.append(_info.__dict__)

        result.print_table_raw(info_dicts, [], auto_header=True)

    else:
        result.print_result_as_table()



"""
FortiAnalyzer get version utility
"""

import logging

from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def version(host: str) -> Result[str]:
    """
    FortiAnalyzer get version

    Args:
        host: Host defined in inventory

    Returns:
        The version string per FAZ

    Raises:
        GeneralWarning: GeneralWarning
    """
    result = Result[str]()
    inventory = Inventory(config.inventory_file)
    assets = inventory.get(host, "fortianalyzer")
    log.debug("FortiAnalyzer get version ...")
    assets[host].login()
    faz_version = assets[host].get_version()
    assets[host].logout()
    result.push_result(host, faz_version)
    return result



"""
The FortiClient EMS get commands
"""

import logging

import typer

from fotoobo.helpers import cli_path
from fotoobo.tools import ems

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The ems get subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def version(
    host: str = typer.Argument(
        "ems",
        help="The FortiClientEMS hostname to access (must be defined in the inventory).",
        metavar="[host]",
    )
) -> None:
    """
    Get the FortiClient EMS version.
    """
    result = ems.get.version(host)
    result.print_result_as_table(
        title="FortiClient EMS Version",
        headers=["FortiClient EMS", "Version"],
    )


@app.command()
def workgroups(
    host: str = typer.Argument(
        "ems",
        help="The FortiClientEMS hostname to access (must be defined in the inventory).",
        metavar="[host]",
    ),
    custom: bool = typer.Option(False, "--custom", "-c", help="Only show custom groups"),
) -> None:
    """
    Get the FortiClient EMS workgroups.
    """
    result = ems.get.workgroups(host, custom)
    result.print_result_as_table(title="FortiClient EMS Workgroups", headers=["Group", "Count"])



"""
Helper functions for testing the cli
"""

from typing import Any, Dict, List, Optional

from rich.text import Text
from rich.tree import Tree


def walk_cli_info(
    info: Dict[str, Any], tree: Tree, command_path: Optional[List[str]] = None
) -> Tree:
    """
    Recursively create the cli command tree from a Typer info dict.

    Args:
        info:         The Typer info dict from Typer.Context.to_info_dict() to walk through
        tree:         The rich tree object to build
        command_path: The command path as a list

    Returns:
        The complete rich Tree object
    """
    command_path = command_path or []
    commands = info["commands"]

    for command in sorted(commands):
        if commands[command]["hidden"]:
            continue

        text = Text(overflow="ellipsis", no_wrap=True)
        text.append(commands[command]["name"], style="bold cyan")
        text.append(" " * max((20 - len(command_path) * 4 - len(command)), 2))
        text.append(commands[command]["help"].split("\n")[0], style="grey53")
        branch = tree.add(text)

        if "commands" in commands[command]:
            walk_cli_info(commands[command], branch, command_path + [command])

    return tree
